from stingray.lightcurve import Lightcurve
from stingray.events import EventList
import numpy as np
from hendrics.io import save_events, HEN_FILE_EXTENSION, load_folding
from hendrics.efsearch import main_zsearch
from hendrics.phaseogram import main_phaseogram, run_interactive_phaseogram
from hendrics.phaseogram import BasePhaseogram
from hendrics.plot import plot_folding
import os
import pytest


class TestPhaseogram():
    def setup_class(cls):
        cls.pulse_frequency = 1/0.101
        cls.tstart = 0
        cls.tend = 25.25
        cls.tseg = cls.tend - cls.tstart
        cls.dt = 0.00606
        cls.times = np.arange(cls.tstart, cls.tend, cls.dt) + cls.dt / 2
        cls.counts = \
            100 + 20 * np.cos(2 * np.pi * cls.times * cls.pulse_frequency)
        lc = Lightcurve(cls.times, cls.counts, gti=[[cls.tstart, cls.tend]])
        events = EventList()
        events.simulate_times(lc)
        events.mjdref = 57000.
        cls.event_times = events.time
        cls.dum = 'events' + HEN_FILE_EXTENSION
        save_events(events, cls.dum)

    def test_zsearch(self):
        evfile = self.dum
        main_zsearch([evfile, '-f', '9.85', '-F', '9.95', '-n', '64',
                      '--fit-candidates', '--fit-frequency',
                      str(self.pulse_frequency)])
        outfile = 'events_Z2n' + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(efperiod.peaks[0], self.pulse_frequency,
                          atol=1/25.25)
        # Defaults to 2 harmonics
        assert efperiod.N == 2

    def test_phaseogram_input_periodogram(self):
        evfile = self.dum
        main_phaseogram([evfile, '--periodogram',
                         'events_Z2n' + HEN_FILE_EXTENSION, '--test'])

    def test_phaseogram_input_f(self):
        evfile = self.dum
        main_phaseogram([evfile, '-f', '9.9', '--test',
                         '--pepoch', '57000'])

    def test_phaseogram_input_f_change(self):
        evfile = self.dum
        ip = run_interactive_phaseogram(evfile, 9.9, test=True, nbin=16, nt=8)
        ip.update(1)
        ip.recalculate(1)
        ip.toa(1)
        ip.reset(1)
        ip.fdot = 2
        f, fdot, fddot = ip.get_values()
        assert fdot == 2
        assert f == 9.9

    def test_phaseogram_raises(self):
        evfile = self.dum
        with pytest.raises(ValueError):
            main_phaseogram([evfile, '--test'])

    def test_phaseogram_input_periodogram_binary(self):
        evfile = self.dum
        main_phaseogram([evfile, '--binary', '--periodogram',
                         'events_Z2n' + HEN_FILE_EXTENSION, '--test',
                         '--pepoch', '57000'])

    def test_phaseogram_input_f_binary(self):
        evfile = self.dum
        main_phaseogram([evfile, '--binary', '-f', '9.9', '--test',
                         '--binary-parameters', '10000', '0', '0'])

    def test_phaseogram_input_f_change_binary(self):
        evfile = self.dum
        ip = run_interactive_phaseogram(evfile, 9.9, test=True, binary=True)
        ip.update(1)
        ip.recalculate(1)
        ip.reset(1)
        ip.zoom_in(1)
        ip.zoom_out(1)
        with pytest.warns(UserWarning) as record:
            ip.toa(1)
        assert np.any(["This function was not implemented" in r.message.args[0]
                       for r in record])
        ip.orbital_period = 2
        orbital_period, fdot, fddot = ip.get_values()
        assert orbital_period == 2

    def test_phaseogram_raises_binary(self):
        evfile = self.dum
        with pytest.raises(ValueError):
            main_phaseogram([evfile, '--binary', '--test'])

    @classmethod
    def teardown_class(cls):
        os.unlink(cls.dum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, print_function
from stingray.events import EventList
import numpy as np
import os
from hendrics.read_events import treat_event_file
from hendrics.io import HEN_FILE_EXTENSION, load_data
from hendrics.base import ref_mjd
import hendrics as hen


class TestReadEvents():
    """Real unit tests."""
    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, 'data')
        cls.fits_fileA = os.path.join(cls.datadir, 'monol_testA.evt')

    def test_treat_event_file_nustar(self):
        treat_event_file(self.fits_fileA)
        new_filename = 'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION
        assert os.path.exists(os.path.join(self.datadir,
                                           new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert 'instr' in data
        assert 'gti' in data
        assert 'mjdref' in data
        assert np.isclose(data['mjdref'], ref_mjd(self.fits_fileA))

    def test_treat_event_file_xmm(self):
        fits_file = os.path.join(self.datadir, 'monol_test_fake.evt')
        hen.fake.main(['--deadtime', '1e-4', '-m', 'XMM', '-i', 'epn',
                       '--ctrate', '2000',
                       '-o', fits_file])

        treat_event_file(fits_file)
        new_filename = 'monol_test_fake_xmm_epn_det01_ev' + HEN_FILE_EXTENSION
        assert os.path.exists(os.path.join(self.datadir,
                                           new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert 'instr' in data
        assert 'gti' in data
        assert 'mjdref' in data
        treat_event_file(fits_file, gti_split=True)
        new_filename = \
            'monol_test_fake_xmm_epn_det01_gti0_ev' + HEN_FILE_EXTENSION
        assert os.path.exists(os.path.join(self.datadir,
                                           new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert 'instr' in data
        assert 'gti' in data
        assert 'mjdref' in data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Quicklook plots."""
from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

try:
    import matplotlib.pyplot as plt
except:
    # Permit to import the module anyway if matplotlib is not present
    pass
from .io import load_data, get_file_type, load_pds, load_lcurve
from .io import is_string, save_as_qdp, load_folding
from stingray.gti import create_gti_mask
from astropy.modeling.models import Const1D
from astropy.modeling import Model

from .base import _assign_value_if_none
from .base import detection_level
import logging
import numpy as np
from .io import HEN_FILE_EXTENSION


def _next_color(ax):
    try:
        return next(ax._get_lines.color_cycle)
    except:
        return next(ax._get_lines.prop_cycler)['color']


def _baseline_fun(x, a):
    """A constant function."""
    return a


def _value_or_none(dict_like, key):
    try:
        return dict_like[key]
    except:
        return None


def plot_generic(fnames, vars, errs=None, figname=None, xlog=None, ylog=None,
                 output_data_file=None):
    """Generic plotting function."""
    if is_string(fnames):
        fnames = [fnames]
    figname = _assign_value_if_none(figname,
                                    '{0} vs {1}'.format(vars[1], vars[0]))
    plt.figure(figname)
    ax = plt.gca()
    if xlog:
        ax.set_xscale('log', nonposx='clip')
    if ylog:
        ax.set_yscale('log', nonposy='clip')

    xlabel, ylabel = vars
    xlabel_err, ylabel_err = None, None
    if errs is not None:
        xlabel_err, ylabel_err = errs

    for i, fname in enumerate(fnames):
        data = load_data(fname)
        color = _next_color(ax)
        xdata = data[xlabel]
        ydata = data[ylabel]
        xdata_err = _value_or_none(data, xlabel_err)
        ydata_err = _value_or_none(data, ylabel_err)
        plt.errorbar(xdata, ydata, yerr=ydata_err, xerr=xdata_err, fmt='-',
                     drawstyle='steps-mid', color=color, label=fname)

        if output_data_file is not None:
            save_as_qdp([xdata, ydata], errors=[xdata_err, ydata_err],
                        filename=output_data_file, mode='a')

    plt.xlabel(vars[0])
    plt.ylabel(vars[1])
    plt.legend()


def _get_const(models):
    """Get constant from Astropy model, list of models or compound model.

    Return None if no Const1D objects are in ``models``.
    Return the value of the first Const1D object found.

    Examples
    --------
    >>> from astropy.modeling.models import Const1D, Gaussian1D
    >>> model = Const1D(2) + Gaussian1D(1, 4, 5)
    >>> _get_const(model)
    2.0
    >>> _get_const(model[0])
    2.0
    >>> _get_const([model[0]])
    2.0
    >>> _get_const([[model]])
    2.0
    >>> _get_const(model[1])

    >>> _get_const(None)

    >>> _get_const(1)

    >>> _get_const('avdsfa')

    """
    import collections

    if isinstance(models, Const1D):
        return models.amplitude.value

    if hasattr(models, 'submodel_names'):
        for subm in models:
            if isinstance(subm, Const1D):
                return subm.amplitude.value

    if models is None:
        return None

    if isinstance(models, collections.Iterable) and \
            not is_string(models) and len(models) != 0:
        for model in models:
            return _get_const(model)

    return None


def plot_pds(fnames, figname=None, xlog=None, ylog=None,
             output_data_file=None, white_sub=False):
    """Plot a list of PDSs, or a single one."""
    from scipy.optimize import curve_fit
    import collections
    if is_string(fnames):
        fnames = [fnames]

    figlabel = fnames[0]

    for i, fname in enumerate(fnames):
        pds_obj = load_pds(fname, nosub=True)
        if np.allclose(np.diff(pds_obj.freq), pds_obj.df):
            freq = pds_obj.freq
            xlog = _assign_value_if_none(xlog, False)
            ylog = _assign_value_if_none(ylog, False)
        else:
            flo = pds_obj.freq - pds_obj.df / 2
            fhi = pds_obj.freq + pds_obj.df / 2
            freq = (fhi + flo) / 2
            xlog = _assign_value_if_none(xlog, True)
            ylog = _assign_value_if_none(ylog, True)

        models = []
        if hasattr(pds_obj, 'best_fits') and pds_obj.best_fits is not None:
            models = pds_obj.best_fits
        if isinstance(models, Model):
            models = [models]

        pds = pds_obj.power
        epds = pds_obj.power_err
        npds = pds_obj.m
        norm = pds_obj.norm

        nbin = len(pds[1:])

        lev = detection_level(nbin, n_summed_spectra=npds)
        if norm == "rms":
            # we need the unnormalized power
            lev = lev / 2 * pds_obj.nphots
            lev, _ = pds_obj._normalize_crossspectrum(lev, pds_obj.fftlen)

        if xlog and ylog:
            plt.figure('PDS - Loglog ' + figlabel)
        else:
            plt.figure('PDS ' + figlabel)
        ax = plt.gca()
        color = _next_color(ax)

        if xlog:
            ax.set_xscale('log', nonposx='clip')
        if ylog:
            ax.set_yscale('log', nonposy='clip')

        level = lev  # Can be modified below
        y = pds[1:]
        yerr = epds[1:]

        if norm.lower() == 'leahy' or (norm.lower() in ['rms', 'frac'] and
                                       (not xlog or not ylog)):
            plt.errorbar(freq[1:], y, yerr=yerr, fmt='-',
                         drawstyle='steps-mid', color=color, label=fname)
            for i, func in enumerate(models):
                plt.plot(freq, func(freq),
                         label='Model {}'.format(i + 1), zorder=20, color='k')

        elif norm.lower() in ['rms', 'frac'] and xlog and ylog:
            # TODO: Very rough! Use new machinery
            const = _get_const(models)
            if const is None:
                p, pcov = curve_fit(_baseline_fun, freq, pds, p0=[2],
                                    sigma=epds)
                logging.info('White noise level is {0}'.format(p[0]))
                const = p[0]

            pds -= const
            level = lev - const

            y = pds[1:] * freq[1:]
            yerr = epds[1:] * freq[1:]
            plt.errorbar(freq[1:], y,
                         yerr=yerr, fmt='-',
                         drawstyle='steps-mid', color=color, label=fname)
            level *= freq
            for i, func in enumerate(models):
                const = _get_const(func)
                plt.plot(freq, freq*(func(freq) - const),
                         label='Model {}'.format(i + 1), zorder=20, color='k')

        if np.any(level < 0):
            continue
        if isinstance(level, collections.Iterable):
            plt.plot(freq, level, ls='--', color=color)
        else:
            plt.axhline(level, ls='--', color=color)
        if output_data_file is not None:
            save_as_qdp([freq[1:], y], errors=[None, yerr],
                        filename=output_data_file, mode='a')

    plt.xlabel('Frequency')
    if norm.lower() == 'rms':
        plt.ylabel('(rms/mean)^2')
    elif norm.lower() == 'leahy':
        plt.ylabel('Leahy power')

    plt.legend()

    if figname is not None:
        plt.savefig(figname)


def plot_cospectrum(fnames, figname=None, xlog=None, ylog=None,
                    output_data_file=None):
    """Plot the cospectra from a list of CPDSs, or a single one."""
    if is_string(fnames):
        fnames = [fnames]

    figlabel = fnames[0]

    for i, fname in enumerate(fnames):
        pds_obj = load_pds(fname, nosub=True)
        models = []
        if hasattr(pds_obj, 'best_fits') and pds_obj.best_fits is not None:
            models = pds_obj.best_fits

        if np.allclose(np.diff(pds_obj.freq), pds_obj.df):
            freq = pds_obj.freq
            xlog = _assign_value_if_none(xlog, False)
            ylog = _assign_value_if_none(ylog, False)
        else:
            flo = pds_obj.freq - pds_obj.df / 2
            fhi = pds_obj.freq + pds_obj.df / 2
            freq = (fhi + flo) / 2
            xlog = _assign_value_if_none(xlog, True)
            ylog = _assign_value_if_none(ylog, True)

        cpds = pds_obj.power

        cospectrum = cpds.real
        if xlog and ylog:
            plt.figure('Cospectrum - Loglog ' + figlabel)
        else:
            plt.figure('Cospectrum ' + figlabel)
        ax = plt.gca()
        if xlog:
            ax.set_xscale('log', nonposx='clip')
        if ylog:
            ax.set_yscale('log', nonposy='clip')

        plt.xlabel('Frequency')
        if xlog and ylog:
            y = freq[1:] * cospectrum[1:]
            plt.plot(freq[1:], y,
                     drawstyle='steps-mid', label=fname)
            for i, func in enumerate(models):
                plt.plot(freq, freq * func(freq),
                         label='Model {}'.format(i + 1))

            plt.ylabel('Cospectrum * Frequency')
        else:
            y = cospectrum[1:]
            plt.plot(freq[1:], cospectrum[1:], drawstyle='steps-mid',
                     label=fname)

            plt.ylabel('Cospectrum')
            for i, func in enumerate(models):
                plt.plot(freq, func(freq),
                         label='Model {}'.format(i + 1))
        if output_data_file is not None:
            save_as_qdp([freq[1:], y], filename=output_data_file, mode='a')

    plt.legend()

    if figname is not None:
        plt.savefig(figname)


def plot_folding(fnames, figname=None, xlog=None, ylog=None,
                 output_data_file=None):
    if is_string(fnames):
        fnames = [fnames]

    for fname in fnames:
        ef = load_folding(fname)

        if len(ef.stat.shape) > 1 and ef.stat.shape[0] > 1:
            plt.figure(fname)
            plt.pcolormesh(ef.freq, np.asarray(ef.fdots), ef.stat)
            plt.colorbar()
            plt.xlabel('Frequency (Hz)')
            plt.ylabel('Fdot (Hz/s)')
        else:
            plt.plot(ef.freq, ef.stat, drawstyle='steps-mid', label=fname)
            plt.xlabel('Frequency (Hz)')
            plt.ylabel(ef.kind + ' stat')

        if hasattr(ef, 'best_fits') and ef.best_fits is not None and \
                not len(ef.stat.shape) > 1:

            for f in ef.best_fits:
                xs = np.linspace(np.min(ef.freq), np.max(ef.freq),
                                 len(ef.freq)*2)
                plt.plot(xs, f(xs))

        if output_data_file is not None:
            out = [ef.freq.flatten(), ef.fdots.flatten(), ef.stat.flatten()]
            out_err = [None, None, None]

            if hasattr(ef, 'best_fits') and ef.best_fits is not None and \
                    not len(ef.stat.shape) > 1:
                for f in ef.best_fits:
                    out.append(f(ef.freq.flatten()))
                    out_err.append(None)

            save_as_qdp(out, out_err, filename=output_data_file, mode='a')
    ax = plt.gca()
    if xlog:
        ax.set_xscale('log', nonposx='clip')
    if ylog:
        ax.set_yscale('log', nonposy='clip')
    plt.legend()

    if figname is not None:
        plt.savefig(figname)


def plot_color(file0, file1, xlog=None, ylog=None, figname=None,
               output_data_file=None):
    type0, lc0 = get_file_type(file0)
    type1, lc1 = get_file_type(file1)
    xlabel, ylabel = 'Count rate', 'Count rate'
    if type0 == 'color':
        xlabel = '{3}-{2}/{1}-{0}'.format(*lc0.e_intervals)
    if type1 == 'color':
        ylabel = '{3}-{2}/{1}-{0}'.format(*lc1.e_intervals)
    plt.errorbar(lc0.counts, lc1.counts,
                 xerr=lc0.counts_err, yerr=lc1.counts_err, fmt='o',
                 color='k', alpha=0.5)
    plt.scatter(lc0.counts, lc1.counts, zorder=10)

    if output_data_file is not None:
        save_as_qdp([lc0.counts, lc1.counts],
                    errors=[lc0.counts_err, lc1.counts_err],
                    filename=output_data_file,
                    mode='a')

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    ax = plt.gca()
    if xlog:
        ax.set_xscale('log', nonposx='clip')
    if ylog:
        ax.set_yscale('log', nonposy='clip')
    if figname is not None:
        plt.savefig(figname)


def plot_lc(lcfiles, figname=None, fromstart=False, xlog=None, ylog=None,
            output_data_file=None):
    """Plot a list of light curve files, or a single one."""
    if is_string(lcfiles):
        lcfiles = [lcfiles]

    figlabel = lcfiles[0]

    plt.figure('LC ' + figlabel)
    for lcfile in lcfiles:
        logging.info('Loading %s...' % lcfile)
        lcdata = load_data(lcfile)

        time = lcdata['time']
        lc = lcdata['counts']
        gti = lcdata['gti']
        instr = lcdata['instr']

        if fromstart:
            time -= lcdata['Tstart']
            gti -= lcdata['Tstart']

        if instr == 'PCA':
            # If RXTE, plot per PCU count rate
            npcus = lcdata['nPCUs']
            lc /= npcus

        for g in gti:
            plt.axvline(g[0], ls='-', color='red')
            plt.axvline(g[1], ls='--', color='red')

        good = create_gti_mask(time, gti)
        plt.plot(time, lc, drawstyle='steps-mid', color='grey')
        plt.plot(time[good], lc[good], drawstyle='steps-mid', label=lcfile)
        if 'base' in lcdata:
            plt.plot(time, lcdata['base'], color='r')

        if output_data_file is not None:
            outqdpdata = [time[good], lc[good]]
            if 'base' in lcdata:
                outqdpdata.append(lcdata['base'][good])
            save_as_qdp(outqdpdata, filename=output_data_file,
                        mode='a')

    plt.xlabel('Time (s)')
    if instr == 'PCA':
        plt.ylabel('light curve (Ct/bin/PCU)')
    else:
        plt.ylabel('light curve (Ct/bin)')

    plt.legend()
    if figname is not None:
        plt.savefig(figname)


def main(args=None):
    """Main function called by the `HENplot` command line script."""
    import argparse

    description = \
        'Plot the content of HENDRICS light curves and frequency spectra'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("--noplot", help="Only create images, do not plot",
                        default=False, action='store_true')
    parser.add_argument("--CCD",
                        help="This is a color-color diagram. In this case, the"
                             " list of files is expected to be given as "
                             "soft0.nc, hard0.nc, soft1.nc, hard1.nc, ...",
                        default=False, action='store_true')
    parser.add_argument("--HID",
                        help="This is a hardness-intensity diagram. In this "
                             "case, the list of files is expected to be given "
                             "as color0.nc, intensity0.nc, color1.nc, "
                             "intensity1.nc, ...",
                        default=False, action='store_true')
    parser.add_argument("--figname", help="Figure name",
                        default=None, type=str)
    parser.add_argument('-o', "--outfile",
                        help="Output data file in QDP format",
                        default=None, type=str)
    parser.add_argument("--xlog", help="Use logarithmic X axis",
                        default=None, action='store_true')
    parser.add_argument("--ylog", help="Use logarithmic Y axis",
                        default=None, action='store_true')
    parser.add_argument("--xlin", help="Use linear X axis",
                        default=None, action='store_true')
    parser.add_argument("--ylin", help="Use linear Y axis",
                        default=None, action='store_true')
    parser.add_argument("--fromstart",
                        help="Times are measured from the start of the "
                             "observation (only relevant for light curves)",
                        default=False, action='store_true')
    parser.add_argument("--axes", nargs=2, type=str,
                        help="Plot two variables contained in the file",
                        default=None)

    args = parser.parse_args(args)
    if args.noplot and args.figname is None:
        args.figname = args.files[0].replace(HEN_FILE_EXTENSION, '.png')

    if args.xlin is not None:
        args.xlog = False
    if args.ylin is not None:
        args.ylog = False

    if args.CCD or args.HID:
        args.files = zip(args.files[:-1:2], args.files[1::2])

    for fname in args.files:

        if args.CCD or args.HID:
            plot_color(fname[0], fname[1], xlog=args.xlog, ylog=args.ylog,
                       figname=args.figname, output_data_file=args.outfile)
            continue
        ftype, contents = get_file_type(fname)
        if args.axes is not None:
            plot_generic(fname, args.axes, xlog=args.xlog, ylog=args.ylog,
                         figname=args.figname, output_data_file=args.outfile)
            continue
        if ftype == 'lc':
            plot_lc(fname, fromstart=args.fromstart, xlog=args.xlog,
                    ylog=args.ylog, figname=args.figname,
                    output_data_file=args.outfile)
        elif ftype == 'folding':
            plot_folding(fname, xlog=args.xlog, ylog=args.ylog,
                         figname=args.figname, output_data_file=args.outfile)
        elif ftype[-4:] == 'cpds':
            plot_cospectrum(fname, xlog=args.xlog, ylog=args.ylog,
                            figname=args.figname,
                            output_data_file=args.outfile)
        elif ftype[-3:] == 'pds':
            plot_pds(fname, xlog=args.xlog, ylog=args.ylog,
                     figname=args.figname, output_data_file=args.outfile)

    if not args.noplot:
        plt.show()



from stingray.lightcurve import Lightcurve
from stingray.events import EventList
import numpy as np
from hendrics.io import save_events, HEN_FILE_EXTENSION, load_folding, \
    load_events
from hendrics.efsearch import main_efsearch, main_zsearch
from hendrics.efsearch import decide_binary_parameters, folding_orbital_search
from hendrics.fold import main_fold
from hendrics.plot import plot_folding
import os
import pytest
try:
    import pandas as pd
    HAS_PD = True
except:
    HAS_PD = False


class TestEFsearch():
    def setup_class(cls):
        cls.pulse_frequency = 1/0.101
        cls.tstart = 0
        cls.tend = 25.25
        cls.tseg = cls.tend - cls.tstart
        cls.dt = 0.00606
        cls.times = np.arange(cls.tstart, cls.tend, cls.dt) + cls.dt / 2
        cls.counts = \
            100 + 20 * np.cos(2 * np.pi * cls.times * cls.pulse_frequency)
        lc = Lightcurve(cls.times, cls.counts, gti=[[cls.tstart, cls.tend]])
        events = EventList()
        events.simulate_times(lc)
        cls.event_times = events.time
        cls.dum_noe = 'events_noe' + HEN_FILE_EXTENSION
        save_events(events, cls.dum_noe)
        events.pi = np.random.uniform(0, 1000, len(events.time))
        cls.dum_pi = 'events_pi' + HEN_FILE_EXTENSION
        save_events(events, cls.dum_pi)
        events.energy = np.random.uniform(3, 79, len(events.time))
        cls.dum = 'events' + HEN_FILE_EXTENSION
        save_events(events, cls.dum)

    def test_fold(self):
        evfile = self.dum
        evfile_noe = self.dum_noe
        evfile_pi = self.dum_pi

        main_fold([evfile, '-f', str(self.pulse_frequency), '-n', '64',
                   '--test', '--norm', 'ratios'])
        outfile = 'Energyprofile_ratios.png'
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold([evfile, '-f', str(self.pulse_frequency), '-n', '64',
                   '--test', '--norm', 'to1'])
        outfile = 'Energyprofile_to1.png'
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold([evfile, '-f', str(self.pulse_frequency), '-n', '64',
                   '--test', '--norm', 'blablabla'])
        outfile = 'Energyprofile_to1.png'
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold([evfile_pi, '-f', str(self.pulse_frequency), '-n', '64',
                   '--test', '--norm', 'blablabla'])
        outfile = 'Energyprofile_to1.png'
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold([evfile_noe, '-f', str(self.pulse_frequency), '-n', '64',
                   '--test', '--norm', 'blablabla'])
        outfile = 'Energyprofile.png'
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_efsearch(self):
        evfile = self.dum
        main_efsearch([evfile, '-f', '9.85', '-F', '9.95', '-n', '64',
                       '--fit-candidates'])
        outfile = 'events_EF' + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(efperiod.peaks[0], self.pulse_frequency,
                          atol=1/25.25)
        os.unlink(outfile)

    def test_zsearch(self):
        evfile = self.dum
        main_zsearch([evfile, '-f', '9.85', '-F', '9.95', '-n', '64',
                      '--fit-candidates', '--fit-frequency',
                      str(self.pulse_frequency),
                      '--dynstep', '5'])
        outfile = 'events_Z2n' + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(efperiod.peaks[0], self.pulse_frequency,
                          atol=1/25.25)
        # Defaults to 2 harmonics
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fdots(self):
        evfile = self.dum
        main_zsearch([evfile, '-f', '9.85', '-F', '9.95', '-n', '64',
                      '--fdotmin', ' -0.1', '--fdotmax', '0.1',
                      '--fit-candidates', '--fit-frequency',
                      str(self.pulse_frequency)])
        outfile = 'events_Z2n' + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file='bla.qdp')
        efperiod = load_folding(outfile)
        assert np.isclose(efperiod.peaks[0], self.pulse_frequency,
                          atol=1/25.25)
        # Defaults to 2 harmonics
        assert len(efperiod.fdots) > 1
        assert efperiod.N == 2
        # os.unlink(outfile)

    @pytest.mark.skipif('not HAS_PD')
    def test_orbital(self):
        import pandas as pd
        events = load_events(self.dum)
        csv_file = decide_binary_parameters(137430, [0.03, 0.035],
                                            [2. * 86400, 2.5 * 86400],
                                            [0., 1.],
                                            fdot_range=[0, 5e-10], reset=False,
                                            NMAX=10)
        table = pd.read_csv(csv_file)
        assert len(table) == 10
        folding_orbital_search(events, csv_file, chunksize=10,
                               outfile='out.csv')
        table = pd.read_csv('out.csv')
        assert len(table) == 10
        assert np.all(table['done'])
        os.unlink(csv_file)
        os.unlink('out.csv')

    @classmethod
    def teardown_class(cls):
        os.unlink(cls.dum)



# -*- coding: utf-8 -*-
"""
@author: marta
"""
from __future__ import print_function, division
import logging
import numpy as np
from .base import hen_root
from .io import load_events
from .io import save_as_qdp
from stingray.varenergyspectrum import RmsEnergySpectrum
from stingray.varenergyspectrum import LagEnergySpectrum
# from stingray.covariancespectrum import AveragedCovariancespectrum


def main(args=None):
    import argparse
    description = ('Calculates variability-energy spectra')
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument('-f', "--freq-interval", nargs=2, type=float,
                        default=[0., 100], help="Frequence interval")
    parser.add_argument("--energy-values", nargs=4, type=str,
                        default="0.3 12 5 lin".split(" "),
                        help="Choose Emin, Emax, number of intervals,"
                        "interval spacing, lin or log")
    parser.add_argument("--segment-size", type=float,
                        default=512,
                        help="Length of the light curve intervals to be "
                             "averaged")
    parser.add_argument("-b", "--bin-time", type=float,
                        default=1,
                        help="Bin time for the light curve")
    parser.add_argument("--ref-band", nargs=2, type=float,
                        default=None, help="Reference band when relevant")
    parser.add_argument("--rms", default=False, action='store_true',
                        help="Calculate rms")
    parser.add_argument("--covariance", default=False, action='store_true',
                        help="Calculate covariance spectrum")
    parser.add_argument("--use-pi", default=False, action='store_true',
                        help="Energy intervals are specified as PI channels")
    parser.add_argument("--cross-instr", default=False, action='store_true',
                        help="Use data files in pairs, for example with the"
                             "reference band from one and the subbands from "
                             "the  other (useful in NuSTAR and "
                             "multiple-detector missions)")
    parser.add_argument("--lag", default=False, action='store_true',
                        help="Calculate lag-energy")
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING', type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENvarenergy.log', level=numeric_level,
                        filemode='w')
    filelist = []
    energy_spec = (float(args.energy_values[0]),
                   float(args.energy_values[1]),
                   int(args.energy_values[2]),
                   args.energy_values[3])

    from .io import sort_files
    if args.cross_instr:
        logging.info('Sorting file list')
        sorted_files = sort_files(args.files)

        logging.warning('Beware! For cpds and derivatives, I assume that the '
                        'files are from only two instruments and in pairs '
                        '(even in random order)')

        instrs = list(sorted_files.keys())

        files1 = sorted_files[instrs[0]]
        files2 = sorted_files[instrs[1]]
    else:
        files1 = args.files
        files2 = args.files

    for fnames in zip(files1, files2):
        fname = fnames[0]
        fname2 = fnames[1]

        events = load_events(fname)
        events2 = load_events(fname2)
        if not args.use_pi and \
                (events.energy is None or events2.energy is None):
            raise ValueError("If --use-pi is not specified, event lists must "
                             "be calibrated! Please use HENcalibrate.")

        if args.rms:
            rms = RmsEnergySpectrum(events, args.freq_interval,
                                    energy_spec,
                                    segment_size=args.segment_size,
                                    bin_time=args.bin_time,
                                    events2=events2,
                                    use_pi=args.use_pi)
            out1 = hen_root(fname) + "_rms" + '.qdp'
            start_energy = np.asarray(rms.energy_intervals)[:, 0]
            stop_energy = np.asarray(rms.energy_intervals)[:, 1]
            save_as_qdp([start_energy, stop_energy, rms.spectrum],
                        [None, None, rms.spectrum_error], filename=out1)
            filelist.append(out1)

        if args.lag:
            lag = LagEnergySpectrum(events, args.freq_interval,
                                    energy_spec, args.ref_band,
                                    segment_size=args.segment_size,
                                    bin_time=args.bin_time,
                                    events2=events2,
                                    use_pi=args.use_pi)
            start_energy = np.asarray(lag.energy_intervals)[:, 0]
            stop_energy = np.asarray(lag.energy_intervals)[:, 1]
            out2 = hen_root(fname) + "_lag" + '.qdp'
            save_as_qdp([start_energy, stop_energy, lag.spectrum],
                        [None, None, lag.spectrum_error], filename=out2)
            filelist.append(out2)

        if args.covariance:
            try:
                from stingray.varenergyspectrum import CovarianceEnergySpectrum
            except:
                logging.warning('This version of Stingray does not implement '
                                'the correct version of Covariance Spectrum.')
                continue
            cov = CovarianceEnergySpectrum(events, args.freq_interval,
                                           energy_spec, args.ref_band,
                                           segment_size=args.segment_size,
                                           bin_time=args.bin_time,
                                           events2=events2,
                                           use_pi=args.use_pi)
            start_energy = np.asarray(cov.energy_intervals)[:, 0]
            stop_energy = np.asarray(cov.energy_intervals)[:, 1]
            out2 = hen_root(fname) + "_cov" + '.qdp'
            save_as_qdp([start_energy, stop_energy, cov.spectrum],
                        [None, None, cov.spectrum_error], filename=out2)
            filelist.append(out2)

    return filelist



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to save data in a Xspec-readable format."""
from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from .io import get_file_type
import numpy as np
from .io import get_file_extension
import subprocess as sp
import logging


def save_as_xspec(fname, direct_save=False, save_lags=True):
    """Save frequency spectra in a format readable to FTOOLS and Xspec.

    Parameters
    ----------
    fname : str
        Input HENDRICS frequency spectrum file name
    direct_save : bool
        If True, also call `flx2xsp` to produce the output .pha and .rsp files.
        If False (default), flx2xsp has to be called from the user

    Notes
    -----
    Uses method described here:
    https://asd.gsfc.nasa.gov/XSPECwiki/fitting_timing_power_spectra_in_XSPEC
    """
    ftype, contents = get_file_type(fname)

    outroot = fname.replace(get_file_extension(fname), '')
    outname = outroot + '_xsp.dat'
    outroot_lags = outroot + '_lags'
    outname_lags = outroot_lags + '_xsp.dat'

    if ftype.endswith('pds'):
        flo = contents.freq - contents.df / 2
        fhi = contents.freq + contents.df / 2
        power = contents.power.real * contents.df
        power_err = contents.power_err.real * contents.df
    else:
        raise ValueError("Data type not supported for Xspec")

    np.savetxt(outname, np.transpose([flo, fhi, power, power_err]))
    if direct_save:
        sp.check_call('flx2xsp {0} {1}.pha {1}.rsp'.format(
            outname, outroot).split())

    if save_lags and ftype == 'cpds':
        lags, lags_err = contents.time_lag()
        np.savetxt(outname_lags, np.transpose([flo, fhi,
                                               lags * contents.df,
                                               lags_err * contents.df]))
        if direct_save:
            sp.check_call('flx2xsp {0} {1}.pha {1}.rsp'.format(
                outname_lags, outroot_lags).split())


def main(args=None):
    """Main function called by the `HEN2xspec` command line script."""
    import argparse
    description = ('Save a frequency spectrum in a qdp file that can be '
                   'read by flx2xsp and produce a XSpec-compatible spectrum'
                   'file')
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("--flx2xsp", help="Also call flx2xsp at the end",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files
    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HEN2xpec.log', level=numeric_level,
                        filemode='w')

    for f in files:
        save_as_xspec(f, direct_save=args.flx2xsp)



# this contains imports plugins that configure py.test for astropy tests.
# by importing them here in conftest.py they are discoverable by py.test
# no matter how it is invoked within the source tree.
"""Configuration for py.test."""

from astropy.tests.pytest_plugins import *   # NOQA

# # Uncomment the following line to treat all DeprecationWarnings as
# # exceptions
# enable_deprecations_as_exceptions()

# # Uncomment and customize the following lines to add/remove entries
# # from the list of packages for which version numbers are displayed
# # when running the tests
# try:
#     PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
#     PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
#     del PYTEST_HEADER_MODULES['h5py']
# except NameError:  # needed to support Astropy < 1.0
#     pass

# Uncomment the following lines to display the version number of the
# package rather than the version number of Astropy in the top line when
# running the tests.
import os

# This is to figure out the affiliated package version, rather than
# using Astropy's
from . import version

try:
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version.version
except NameError:   # Needed to support Astropy <= 1.0.0
    pass


try:
    import matplotlib
except ImportError:
    pass
else:
    matplotlib.use('Agg')



#!C:\bld\hendrics_1543255133425\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'hendrics==4.0rc1.post2','console_scripts','HENlags'
__requires__ = 'hendrics==4.0rc1.post2'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('hendrics==4.0rc1.post2', 'console_scripts', 'HENlags')()
    )



import os
from astropy.io import fits
import numpy as np
import pytest
from ..phasetag import main_phasetag
from ..read_events import main as main_read
from ..io import load_events, HEN_FILE_EXTENSION, load_events_and_gtis


class TestPhasetag():
    """Real unit tests."""
    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, 'data')
        cls.fits_fileA = os.path.join(cls.datadir, 'monol_testA.evt')
        cls.freq = 0.1235242

    @pytest.mark.parametrize('N', [5, 6, 11, 16, 32, 41])
    def test_phase_tag(self, N):
        main_phasetag([self.fits_fileA, '-f', str(self.freq), '--test',
                       '--tomax', '-n', str(N), '--plot'])
        self.phasetagged = self.fits_fileA.replace('.evt', '_phasetag.evt')
        assert os.path.exists(self.phasetagged)

        # Redo to test if existing columns are preserved
        main_phasetag([self.phasetagged, '-f', str(self.freq), '--test',
                       '--tomax', '-n', str(N)])

        hdulist = fits.open(self.phasetagged)

        times = np.array(hdulist[1].data['TIME'])
        phases = np.array(hdulist[1].data['Phase'])
        orbtimes = np.array(hdulist[1].data['Orbit_bary'])
        hdulist.close()

        assert np.all(phases >= 0)
        assert np.all(phases < 1)
        # There's no orbit in these files
        assert np.allclose(times, orbtimes, rtol=1e-15)

        phase_bins = np.linspace(0, 1, N + 1)
        prof, _ = np.histogram(phases, bins=phase_bins)

        # I used --tomax
        assert np.argmax(prof) == 0
        os.unlink(self.phasetagged)

    @pytest.mark.parametrize('N', [5, 6, 11, 16, 32, 41])
    def test_phase_tag_TOA(self, N):
        main_phasetag([self.fits_fileA, '-f', str(self.freq), '--test',
                       '--refTOA', '57000.01', '-n', str(N), '--plot'])
        self.phasetagged = self.fits_fileA.replace('.evt', '_phasetag.evt')
        assert os.path.exists(self.phasetagged)

    def test_phase_tag_badexposure(self):
        with pytest.warns(UserWarning) as record:
            main_phasetag([self.fits_fileA, '-f', '0.00001', '--test',
                           '--tomax', '-n', '1000', '--plot'])
        assert np.any(["Exposure has NaNs or zeros. " in r.message.args[0]
                       for r in record])

    def test_phase_tag_invalid0(self):
        with pytest.raises(ValueError) as excinfo:
            main_phasetag([self.fits_fileA, '--test'])
        assert 'Specify one between' in str(excinfo)

    def test_phase_tag_invalid1(self):
        with pytest.raises(ValueError) as excinfo:
            main_phasetag([self.fits_fileA, '-f', '1', '--parfile', 'bubu.par',
                           '--test'])
        assert 'Specify only one between' in str(excinfo)

    def test_phase_tag_parfile(self):
        with pytest.raises(NotImplementedError) as excinfo:
            main_phasetag([self.fits_fileA, '--parfile', 'bubu.par',
                           '--test'])
        assert 'This part is not yet implemented' in str(excinfo)



"""Save different input files in PRESTO-readable format."""
from __future__ import print_function, division

import logging
from astropy.coordinates import SkyCoord
import numpy as np
from .io import high_precision_keyword_read, get_file_type, HEN_FILE_EXTENSION

MAXBIN = 100000000


def get_header_info(obj):
    """Get header info from a Stingray object."""
    from astropy.io.fits import Header
    header = Header.fromstring(obj.header)
    info = type('', (), {})()
    info.mjdref = high_precision_keyword_read(header, 'MJDREF')
    info.telescope = header['TELESCOP']
    info.instrument = header['INSTRUME']
    info.source = header['OBJECT']
    try:
        user = header['USER']
    except:
        user = 'Unknown'
    info.observer = user
    info.user = user
    info.tstart = header['TSTART']
    info.tstop = header['TSTOP']
    try:
        ra = header['RA_OBJ']
        dec = header['DEC_OBJ']
    except:
        ra = header['RA_PNT']
        dec = header['DEC_PNT']

    a = SkyCoord(ra, dec, unit="degree")
    info.raj = \
        (a.ra.to_string("hourangle")
         ).replace("s", "").replace("h", ":").replace("m", ":")
    info.decj = (a.ra.to_string()
                 ).replace("s", "").replace("d", ":").replace("m", ":")
    if hasattr(obj, 'e_interval'):
        e0, e1 = obj.e_interval
    elif hasattr(obj, 'energy') and obj.energy is not None:
        e0, e1 = np.min(obj.energy), np.max(obj.energy)
    else:
        e0, e1 = 0, 0
    info.centralE = (e0 + e1) / 2
    info.bandpass = e1 - e0

    return info


def _save_to_binary(lc, filename):
    """Save a light curve to binary format."""
    nc = len(lc.counts)
    lc.counts[:nc // 2 * 2].astype('float32').tofile(filename)
    return


def save_lc_to_binary(lc, filename):
    """Save a light curve to binary format.

    Parameters
    ----------
    lc : `:class:stingray.Lightcurve`
        Input light curve
    filename : str
        Output file name
    Returns
    -------
    lcinfo : object
        light curve info
    """
    tstart = lc.tstart
    tstop = lc.tstart + lc.tseg
    nbin = lc.n
    bin_time = lc.dt
    _save_to_binary(lc, filename)

    lcinfo = type('', (), {})()
    lcinfo.bin_intervals_start = [0]
    lcinfo.bin_intervals_stop = [nbin]
    lcinfo.lclen = nbin
    lcinfo.tstart = tstart
    lcinfo.dt = bin_time
    lcinfo.tseg = tstop - tstart
    lcinfo.nphot = np.sum(lc.counts)
    return lcinfo


def save_events_to_binary(events, filename, bin_time, tstart=None,
                          emin=None, emax=None):
    """Save an event list to binary format.

    Parameters
    ----------
    events : `:class:stingray.Eventlist`
        Input event list
    filename : str
        Output file name
    bin_time : float
        Bin time of the output light curve

    Other parameters
    ----------------
    tstart : float
        Starting time
    emin : float
        Minimum energy of the photons
    emax : float
        Maximum energy of the photons

    Returns
    -------
    lcinfo : object
        light curve info
    """
    import struct
    if tstart is None:
        tstart = events.gti[0, 0]

    if emin is not None and emax is not None:
        if not hasattr(events, 'energy') or events.energy is None:
            raise ValueError(
                'Energy filtering requested for uncalibrated event '
                'list')

        good = (events.energy >= emin) & (events.energy < emax)
        events.time = events.time[good]

    tstop = events.gti[-1, 1]
    nbin = (tstop - tstart) / bin_time

    lclen = 0
    file = open(filename, 'wb')
    nphot = 0
    for i in np.arange(0, nbin, MAXBIN):
        t0 = i * bin_time + tstart

        lastbin = np.min([MAXBIN, (nbin - i) // 2 * 2])
        t1 = t0 + lastbin * bin_time

        good = (events.time >= t0) & (events.time < t1)

        goodev = events.time[good]
        hist, times = \
            np.histogram(goodev, bins=np.linspace(t0, t1, lastbin + 1))

        lclen += lastbin
        s = struct.pack('f'*len(hist), *hist)
        print(s[:8])
        file.write(s)
        nphot += len(goodev)
    file.close()

    lcinfo = type('', (), {})()
    lcinfo.bin_intervals_start = \
        np.floor((events.gti[:, 0] - tstart) / bin_time)
    lcinfo.bin_intervals_stop = \
        np.floor((events.gti[:, 1] - tstart) / bin_time)
    lcinfo.lclen = lclen
    lcinfo.tstart = tstart
    lcinfo.dt = bin_time
    lcinfo.tseg = tstop - tstart
    lcinfo.nphot = nphot
    return lcinfo


def save_inf(lcinfo, info, filename):
    """Save information file."""

    lclen = lcinfo.lclen
    bin_intervals_start, bin_intervals_stop = \
        lcinfo.bin_intervals_start, lcinfo.bin_intervals_stop

    epoch = info.mjdref + lcinfo.tstart / 86400

    with open(filename, 'w') as f:
        print(" Data file name without suffix         "
              " =  {}".format(filename.replace('.inf', '')), file=f)
        print(" Telescope used                        "
              " =  {}".format(info.telescope), file=f)
        print(" Instrument used                       "
              " =  {}".format(info.instrument), file=f)
        print(" Object being observed                 "
              " =  {}".format(info.source), file=f)
        print(" J2000 Right Ascension (hh:mm:ss.ssss) "
              " =  {}".format(info.raj), file=f)
        print(" J2000 Declination     (dd:mm:ss.ssss) "
              " =  {}".format(info.decj), file=f)
        print(" Data observed by                      "
              " =  {}".format(info.observer), file=f)
        print(" Epoch of observation (MJD)            "
              " =  {:05.15f}".format(epoch), file=f)
        print(" Barycentered?           (1=yes, 0=no) "
              " =  1", file=f)
        print(" Number of bins in the time series     "
              " =  {lclen}".format(lclen=lclen), file=f)
        print(" Width of each time series bin (sec)   "
              " =  {bintime}".format(bintime=lcinfo.dt), file=f)
        print(" Any breaks in the data? (1 yes, 0 no) "
              " =  1", file=f)
        for i, st in enumerate(bin_intervals_start):
            print(" On/Off bin pair # {ngti:>2}                  "
                  " =  {binstart:<11}, "
                  "{binstop:<11}".format(ngti=i + 1, binstart=st,
                                         binstop=bin_intervals_stop[i]),
                  file=f)
        print(" Type of observation (EM band)         "
              " =  X-ray", file=f)
        print(" Field-of-view diameter (arcsec)       "
              " =  400", file=f)
        print(" Central energy (kev)                  "
              " =  {}".format(info.centralE), file=f)
        print(" Energy bandpass (kev)                 "
              " =  {}".format(info.bandpass), file=f)
        print(" Data analyzed by                      "
              " =  {}".format(info.user), file=f)
        print(" Any additional notes:", file=f)
        print("       T = {length}, Nphot={nphot}".format(length=lcinfo.tseg,
                                                          nphot=lcinfo.nphot),
              file=f)

    return


def main_presto(args=None):
    import argparse
    from multiprocessing import Pool

    description = ('Save light curves in a format readable to PRESTO')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of input light curves", nargs='+')

    parser.add_argument("-b", "--bin-time", help="Bin time",
                        type=np.longdouble, default=1)
    parser.add_argument("-e", "--energy-interval", help="Energy interval",
                        nargs=2, type=float, default=[None, None])

    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--nproc",
                        help=("Number of processors to use"),
                        default=1,
                        type=int)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENbinary.log', level=numeric_level,
                        filemode='w')

    for f in args.files:
        print(f)
        outfile = f.replace(HEN_FILE_EXTENSION, '.dat')
        ftype, contents = get_file_type(f)
        if ftype == 'lc':
            lcinfo = save_lc_to_binary(contents, outfile)
        elif ftype == 'events':
            lcinfo = save_events_to_binary(contents, outfile,
                                           bin_time=args.bin_time,
                                           emin=args.energy_interval[0],
                                           emax=args.energy_interval[1])
        else:
            raise ValueError('File type not recognized')

        info = get_header_info(contents)
        save_inf(lcinfo, info, f.replace(HEN_FILE_EXTENSION, '.inf'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Read and save event lists from FITS files."""
from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from stingray.utils import assign_value_if_none
from stingray.events import EventList
from .base import hen_root, read_header_key
from .io import save_events, load_events_and_gtis
from .io import HEN_FILE_EXTENSION
import numpy as np
import logging
import warnings
import os


def treat_event_file(filename, noclobber=False, gti_split=False,
                     min_length=4, gtistring=None):
    """Read data from an event file, with no external GTI information.

    Parameters
    ----------
    filename : str

    Other Parameters
    ----------------
    noclobber: bool
        if a file is present, do not overwrite it
    gtistring: str
        comma-separated set of GTI strings to consider
    gti_split: bool
        split the file in multiple chunks, containing one GTI each
    min_length: float
        minimum length of GTIs accepted (only if gti_split is True)
    """
    gtistring = assign_value_if_none(gtistring, 'GTI,STDGTI')
    logging.info('Opening %s' % filename)

    instr = read_header_key(filename, 'INSTRUME')
    mission = read_header_key(filename, 'TELESCOP')

    data = load_events_and_gtis(filename,
                                gtistring=gtistring)

    events = data.ev_list
    gtis = events.gti
    detector_id = data.detector_id

    if detector_id is not None:
        detectors = np.array(list(set(detector_id)))
    else:
        detectors = [None]
    outfile_root = \
        hen_root(filename) + '_' + mission.lower() + '_' + instr.lower()

    for d in detectors:
        if d is not None:
            good_det = d == data.detector_id
            outroot_local = \
                '{0}_det{1:02d}'.format(outfile_root, d)

        else:
            good_det = np.ones_like(events.time, dtype=bool)
            outroot_local = outfile_root

        outfile = outroot_local + '_ev' + HEN_FILE_EXTENSION
        if noclobber and os.path.exists(outfile) and (not gti_split):
            warnings.warn(
                '{0} exists and using noclobber. Skipping'.format(outfile))
            return

        if gti_split:
            for ig, g in enumerate(gtis):
                length = g[1] - g[0]
                if length < min_length:
                    print("GTI shorter than {} s; skipping".format(min_length))
                    continue

                outfile_local = \
                    '{0}_gti{1}_ev'.format(outroot_local,
                                           ig) + HEN_FILE_EXTENSION

                if noclobber and os.path.exists(outfile_local):
                    warnings.warn('{0} exists, '.format(outfile_local) +
                                  'and noclobber option used. Skipping')
                    return
                good = np.logical_and(events.time >= g[0],
                                      events.time < g[1])
                all_good = good_det & good
                events_filt = EventList(events.time[all_good],
                                        pi=events.pi[all_good],
                                        gti=np.array([g], dtype=np.longdouble),
                                        mjdref=events.mjdref)
                events_filt.instr = events.instr
                events_filt.header = events.header
                save_events(events_filt, outfile_local)
            pass
        else:
            events_filt = EventList(events.time[good_det],
                                    pi=events.pi[good_det],
                                    gti=events.gti, mjdref=events.mjdref)
            events_filt.instr = events.instr
            events_filt.header = events.header

            save_events(events_filt, outfile)


def _wrap_fun(arglist):
    f, kwargs = arglist
    return treat_event_file(f, **kwargs)


def main(args=None):
    """Main function called by the `HENreadevents` command line script."""
    import argparse
    from multiprocessing import Pool

    description = ('Read a cleaned event files and saves the relevant '
                   'information in a standard format')
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--nproc",
                        help=("Number of processors to use"),
                        default=1,
                        type=int)
    parser.add_argument("--noclobber",
                        help=("Do not overwrite existing event files"),
                        default=False, action='store_true')
    parser.add_argument("-g", "--gti-split",
                        help="Split event list by GTI",
                        default=False,
                        action="store_true")
    parser.add_argument("--min-length", type=int,
                        help="Minimum length of GTIs to consider",
                        default=0)
    parser.add_argument("--gti-string", type=str,
                        help="GTI string",
                        default=None)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENreadevents.log', level=numeric_level,
                        filemode='w')

    argdict = {"noclobber": args.noclobber, "gti_split": args.gti_split,
               "min_length": args.min_length, "gtistring": args.gti_string}

    arglist = [[f, argdict] for f in files]

    if os.name == 'nt' or args.nproc == 1:
        [_wrap_fun(a) for a in arglist]
    else:
        pool = Pool(processes=args.nproc)
        for i in pool.imap_unordered(_wrap_fun, arglist):
            pass
        pool.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Calibrate event lists by looking in rmf files."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from .io import load_events, save_events, get_file_extension
from .io import HEN_FILE_EXTENSION
import numpy as np
import os
import logging


def default_nustar_rmf():
    """Look for the default rmf file in the CALDB.

    The CALDB environment variable has to point to the correct location of
    the NuSTAR CALDB

    .. note:: The calibration might change in the future. The hardcoded file
              name will be eventually replaced with a smarter choice based
              on observing time
    """
    logging.warning("Rmf not specified. Using default NuSTAR rmf.")
    rmf = "data/nustar/fpm/cpf/rmf/nuAdet3_20100101v002.rmf"
    path = rmf.split('/')
    newpath = os.path.join(os.environ['CALDB'], *path)
    return newpath


def read_rmf(rmf_file=None):
    """Load RMF info.

    .. note:: Preliminary: only EBOUNDS are read.

    Parameters
    ----------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.

    Returns
    -------
    pis : array-like
        the PI channels
    e_mins : array-like
        the lower energy bound of each PI channel
    e_maxs : array-like
        the upper energy bound of each PI channel
    """
    from astropy.io import fits as pf

    if rmf_file is None or rmf_file == '':
        rmf_file = default_nustar_rmf()

    lchdulist = pf.open(rmf_file, checksum=True)
    lchdulist.verify('warn')
    lctable = lchdulist['EBOUNDS'].data
    pis = np.array(lctable.field('CHANNEL'))
    e_mins = np.array(lctable.field('E_MIN'))
    e_maxs = np.array(lctable.field('E_MAX'))
    lchdulist.close()
    return pis, e_mins, e_maxs


def read_calibration(pis, rmf_file=None):
    """Read the energy channels corresponding to the given PI channels.

    Parameters
    ----------
    pis : array-like
        The channels to lookup in the rmf

    Other Parameters
    ----------------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.
    """
    calp, calEmin, calEmax = read_rmf(rmf_file)
    es = np.zeros(len(pis), dtype=np.float)
    for ic, c in enumerate(calp):
        good = pis == c
        if not np.any(good):
            continue
        es[good] = (calEmin[ic] + calEmax[ic]) / 2

    return es


def calibrate(fname, outname, rmf_file=None):
    """Do calibration of an event list.

    Parameters
    ----------
    fname : str
        The HENDRICS file containing the events
    outname : str
        The output file

    Other Parameters
    ----------------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.
    """
    # Read event file
    logging.info("Loading file %s..." % fname)
    evdata = load_events(fname)
    logging.info("Done.")
    pis = evdata.pi

    es = read_calibration(pis, rmf_file)
    evdata.energy = es
    logging.info('Saving calibrated data to %s' % outname)
    save_events(evdata, outname)


def _calib_wrap(args):
    f, outname, rmf = args
    return calibrate(f, outname, rmf)


def main(args=None):
    """Main function called by the `HENcalibrate` command line script."""
    import argparse
    from multiprocessing import Pool

    description = ('Calibrate clean event files by associating the correct '
                   'energy to each PI channel. Uses either a specified rmf '
                   'file or (for NuSTAR only) an rmf file from the CALDB')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("-r", "--rmf", help="rmf file used for calibration",
                        default=None, type=str)
    parser.add_argument("-o", "--overwrite",
                        help="Overwrite; default: no",
                        default=False,
                        action="store_true")
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("--nproc",
                        help=("Number of processors to use"),
                        default=1,
                        type=int)

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENcalibrate.log', level=numeric_level,
                        filemode='w')

    funcargs = []
    for i_f, f in enumerate(files):
        outname = f
        if args.overwrite is False:
            outname = f.replace(get_file_extension(f), '_calib' +
                                HEN_FILE_EXTENSION)
        funcargs.append([f, outname, args.rmf])

    if os.name == 'nt' or args.nproc == 1:
        [_calib_wrap(fa) for fa in funcargs]
    else:
        pool = Pool(processes=args.nproc)
        for i in pool.imap_unordered(_calib_wrap, funcargs):
            pass
        pool.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to simulate data and produce a fake event file."""

from __future__ import (absolute_import, division,
                        print_function)

import numpy as np
import numpy.random as ra
import os
import logging
import warnings
from stingray.events import EventList
from stingray.lightcurve import Lightcurve
from stingray.utils import assign_value_if_none
from .io import get_file_format, load_data, load_lcurve
from .base import _empty
import copy

from .lcurve import lcurve_from_fits
try:
    from numba import jit
except:
    def jit(fun):
        """Dummy decorator in case jit cannot be imported."""
        return fun


def _paralyzable_dead_time(event_list, dead_time):
    mask = np.ones(len(event_list), dtype=bool)
    dead_time_end = event_list + dead_time
    bad = dead_time_end[:-1] > event_list[1:]
    # Easy: paralyzable case. Here, events coming during dead time produce
    # more dead time. So...
    mask[1:][bad] = False

    return event_list[mask], mask


@jit
def _nonpar_core(event_list, dead_time_end, mask):
    for i in range(1, len(event_list)):
        if (event_list[i] < dead_time_end[i - 1]):
            dead_time_end[i] = dead_time_end[i - 1]
            mask[i] = False
    return mask


def _non_paralyzable_dead_time(event_list, dead_time):
    dead_time_end = event_list + dead_time
    mask = np.ones(len(event_list), dtype=bool)
    mask = _nonpar_core(event_list, dead_time_end, mask)
    return event_list[mask], mask


def filter_for_deadtime(event_list, deadtime, bkg_ev_list=None,
                        dt_sigma=None, paralyzable=False,
                        additional_data=None, return_all=False):
    """Filter an event list for a given dead time.

    Parameters
    ----------
    ev_list : array-like
        The event list
    deadtime: float
        The (mean, if not constant) value of deadtime

    Returns
    -------
    new_ev_list : array-like
        The filtered event list
    additional_output : dict
        Object with all the following attributes. Only returned if
        `return_all` is True
    uf_events : array-like
        Unfiltered event list (events + background)
    is_event : array-like
        Boolean values; True if event, False if background
    deadtime : array-like
        Dead time values
    bkg : array-like
        The filtered background event list
    mask : array-like, optional
        The mask that filters the input event list and produces the output
        event list.

    Other Parameters
    ----------------
    bkg_ev_list : array-like
        A background event list that affects dead time
    dt_sigma : float
        The standard deviation of a non-constant dead time around deadtime.
    return_all : bool
        If True, return the mask that filters the input event list to obtain
        the output event list.

    """
    additional_output = _empty()
    if not isinstance(event_list, EventList):
        event_list_obj = EventList(event_list)
    else:
        event_list_obj = event_list

    ev_list = event_list_obj.time

    if deadtime <= 0.:
        return copy.copy(event_list)

    # Create the total lightcurve, and a "kind" array that keeps track
    # of the events classified as "signal" (True) and "background" (False)
    if bkg_ev_list is not None:
        tot_ev_list = np.append(ev_list, bkg_ev_list)
        ev_kind = np.append(np.ones(len(ev_list), dtype=bool),
                            np.zeros(len(bkg_ev_list), dtype=bool))
        order = np.argsort(tot_ev_list)
        tot_ev_list = tot_ev_list[order]
        ev_kind = ev_kind[order]
        del order
    else:
        tot_ev_list = ev_list
        ev_kind = np.ones(len(ev_list), dtype=bool)

    nevents = len(tot_ev_list)
    all_ev_kind = ev_kind.copy()

    if dt_sigma is not None:
        deadtime_values = ra.normal(deadtime, dt_sigma, nevents)
    else:
        deadtime_values = np.zeros(nevents) + deadtime

    initial_len = len(tot_ev_list)

    if paralyzable:
        tot_ev_list, saved_mask = \
            _paralyzable_dead_time(tot_ev_list, deadtime_values)

    else:
        tot_ev_list, saved_mask = \
            _non_paralyzable_dead_time(tot_ev_list, deadtime_values)

    ev_kind = ev_kind[saved_mask]
    deadtime_values = deadtime_values[saved_mask]
    final_len = len(tot_ev_list)
    logging.info(
        'filter_for_deadtime: '
        '{0}/{1} events rejected'.format(initial_len - final_len,
                                         initial_len))
    retval = EventList(time=tot_ev_list[ev_kind], mjdref=event_list_obj.mjdref)

    if hasattr(event_list_obj, 'pi') and event_list_obj.pi is not None:
        warnings.warn("PI information is lost during dead time filtering")

    if not isinstance(event_list, EventList):
        retval = retval.time

    if return_all:
        additional_output.uf_events = tot_ev_list
        additional_output.is_event = ev_kind
        additional_output.deadtime = deadtime_values
        additional_output.mask = saved_mask[all_ev_kind]
        additional_output.bkg = tot_ev_list[np.logical_not(ev_kind)]
        retval = [retval, additional_output]

    return retval


def generate_fake_fits_observation(event_list=None, filename=None,
                                   instr='FPMA', gti=None, tstart=None,
                                   tstop=None, mission='NUSTAR',
                                   mjdref=55197.00076601852,
                                   livetime=None, additional_columns={}):
    """Generate fake NuSTAR data.

    Takes an event list (as a list of floats)
    All inputs are None by default, and can be set during the call.

    Parameters
    ----------
    event_list : list-like
        :class:`stingray.events.Eventlist` object. If left None, 1000
        random events will be generated, for a total length of 1025 s or the
        difference between tstop and tstart.
    filename : str
        Output file name

    Returns
    -------
    hdulist : FITS hdu list
        FITS hdu list of the output file

    Other Parameters
    ----------------
    mjdref : float
        Reference MJD. Default is 55197.00076601852 (NuSTAR)
    pi : list-like
        The PI channel of each event
    tstart : float
        Start of the observation (s from mjdref)
    tstop : float
        End of the observation (s from mjdref)
    instr : str
        Name of the instrument. Default is 'FPMA'
    livetime : float
        Total livetime. Default is tstop - tstart
    """
    from astropy.io import fits
    import numpy.random as ra

    if event_list is None:
        tstart = assign_value_if_none(tstart, 8e+7)
        tstop = assign_value_if_none(tstop, tstart + 1025)
        ev_list = sorted(ra.uniform(tstart, tstop, 1000))
    else:
        ev_list = event_list.time

    if hasattr(event_list, 'pi'):
        pi = event_list.pi
    else:
        pi = ra.randint(0, 1024, len(ev_list))

    tstart = assign_value_if_none(tstart, np.floor(ev_list[0]))
    tstop = assign_value_if_none(tstop, np.ceil(ev_list[-1]))
    gti = assign_value_if_none(gti, np.array([[tstart, tstop]]))
    filename = assign_value_if_none(filename, 'events.evt')
    livetime = assign_value_if_none(livetime, tstop - tstart)

    if livetime > tstop - tstart:
        raise ValueError('Livetime must be equal or smaller than '
                         'tstop - tstart')

    # Create primary header
    prihdr = fits.Header()
    prihdr['OBSERVER'] = 'Edwige Bubble'
    prihdr['TELESCOP'] = (mission, 'Telescope (mission) name')
    prihdr['INSTRUME'] = (instr, 'Instrument name')
    prihdu = fits.PrimaryHDU(header=prihdr)

    # Write events to table
    col1 = fits.Column(name='TIME', format='1D', array=ev_list)
    col2 = fits.Column(name='PI', format='1J', array=pi)

    allcols = [col1, col2]

    if mission.lower().strip() == 'xmm':
        ccdnr = np.zeros(len(ev_list)) + 1
        ccdnr[1] = 2  # Make it less trivial
        ccdnr[10] = 7
        allcols.append(fits.Column(name='CCDNR', format='1J',
                                   array=ccdnr))

    for c in additional_columns.keys():
        col = fits.Column(name=c, array=additional_columns[c]["data"],
                          format=additional_columns[c]["format"])
        allcols.append(col)

    cols = fits.ColDefs(allcols)
    tbhdu = fits.BinTableHDU.from_columns(cols)
    tbhdu.name = 'EVENTS'

    # ---- Fake lots of information ----
    tbheader = tbhdu.header
    tbheader['OBSERVER'] = 'Edwige Bubble'
    tbheader['COMMENT'] = ("FITS (Flexible Image Transport System) format is"
                           " defined in 'Astronomy and Astrophysics', volume"
                           " 376, page 359; bibcode: 2001A&A...376..359H")
    tbheader['TELESCOP'] = (mission, 'Telescope (mission) name')
    tbheader['INSTRUME'] = (instr, 'Instrument name')
    tbheader['OBS_ID'] = ('00000000001', 'Observation ID')
    tbheader['TARG_ID'] = (0, 'Target ID')
    tbheader['OBJECT'] = ('Fake X-1', 'Name of observed object')
    tbheader['RA_OBJ'] = (0.0, '[deg] R.A. Object')
    tbheader['DEC_OBJ'] = (0.0, '[deg] Dec Object')
    tbheader['RA_NOM'] = (0.0,
                          'Right Ascension used for barycenter corrections')
    tbheader['DEC_NOM'] = (0.0,
                           'Declination used for barycenter corrections')
    tbheader['RA_PNT'] = (0.0, '[deg] RA pointing')
    tbheader['DEC_PNT'] = (0.0, '[deg] Dec pointing')
    tbheader['PA_PNT'] = (0.0, '[deg] Position angle (roll)')
    tbheader['EQUINOX'] = (2.000E+03, 'Equinox of celestial coord system')
    tbheader['RADECSYS'] = ('FK5', 'Coordinate Reference System')
    tbheader['TASSIGN'] = ('SATELLITE', 'Time assigned by onboard clock')
    tbheader['TIMESYS'] = ('TDB', 'All times in this file are TDB')
    tbheader['MJDREFI'] = (int(mjdref),
                           'TDB time reference; Modified Julian Day (int)')
    tbheader['MJDREFF'] = (mjdref - int(mjdref),
                           'TDB time reference; Modified Julian Day (frac)')
    tbheader['TIMEREF'] = ('SOLARSYSTEM',
                           'Times are pathlength-corrected to barycenter')
    tbheader['CLOCKAPP'] = (False, 'TRUE if timestamps corrected by gnd sware')
    tbheader['COMMENT'] = ("MJDREFI+MJDREFF = epoch of Jan 1, 2010, in TT "
                           "time system.")
    tbheader['TIMEUNIT'] = ('s', 'unit for time keywords')
    tbheader['TSTART'] = (tstart,
                          'Elapsed seconds since MJDREF at start of file')
    tbheader['TSTOP'] = (tstop,
                         'Elapsed seconds since MJDREF at end of file')
    tbheader['LIVETIME'] = (livetime, 'On-source time')
    tbheader['TIMEZERO'] = (0.000000E+00, 'Time Zero')
    tbheader['COMMENT'] = (
        "Generated with HENDRICS by {0}".format(os.getenv('USER')))

    # ---- END Fake lots of information ----

    # Fake GTIs

    start = gti[:, 0]
    stop = gti[:, 1]

    col1 = fits.Column(name='START', format='1D', array=start)
    col2 = fits.Column(name='STOP', format='1D', array=stop)
    allcols = [col1, col2]
    cols = fits.ColDefs(allcols)
    gtinames = ['GTI']
    if mission.lower().strip() == 'xmm':
        gtinames = ['STDGTI01', 'STDGTI02', 'STDGTI07']

    all_new_hdus = [prihdu, tbhdu]
    for name in gtinames:
        gtihdu = fits.BinTableHDU.from_columns(cols)
        gtihdu.name = name
        all_new_hdus.append(gtihdu)

    thdulist = fits.HDUList(all_new_hdus)

    thdulist.writeto(filename, overwrite=True)
    return thdulist


def _read_event_list(filename):
    if filename is not None:
        warnings.warn('Input event lists not yet implemented')
    return None, None


def _read_light_curve(filename):
    file_format = get_file_format(filename)
    if file_format == 'fits':
        filename = lcurve_from_fits(filename)[0]
    lc = load_lcurve(filename)

    return lc


def main(args=None):
    """Main function called by the `HENfake` command line script."""
    import argparse
    description = (
        'Create an event file in FITS format from an event list, or simulating'
        ' it. If input event list is not specified, generates the events '
        'randomly')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-e", "--event-list", type=str, default=None,
                        help="File containint event list")
    parser.add_argument("-l", "--lc", type=str, default=None,
                        help="File containing light curve")
    parser.add_argument("-c", "--ctrate", type=float, default=None,
                        help="Count rate for simulated events")
    parser.add_argument("-o", "--outname", type=str, default='events.evt',
                        help="Output file name")
    parser.add_argument("-i", "--instrument", type=str, default='FPMA',
                        help="Instrument name")
    parser.add_argument("-m", "--mission", type=str, default='NUSTAR',
                        help="Mission name")
    parser.add_argument("--tstart", type=float, default=None,
                        help="Start time of the observation (s from MJDREF)")
    parser.add_argument("--tstop", type=float, default=None,
                        help="End time of the observation (s from MJDREF)")
    parser.add_argument("--mjdref", type=float, default=55197.00076601852,
                        help="Reference MJD")
    parser.add_argument("--deadtime", type=float, default=None, nargs='+',
                        help="Dead time magnitude. Can be specified as a "
                             "single number, or two. In this last case, the "
                             "second value is used as sigma of the dead time "
                             "distribution")

    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENfake.log', level=numeric_level,
                        filemode='w')

    additional_columns = {}
    livetime = None
    if args.lc is None and args.ctrate is None and args.event_list is not None:
        event_list = _read_event_list(args.event_list)
    elif args.lc is not None or args.ctrate is not None:
        event_list = EventList()
        if args.lc is not None:
            lc = _read_light_curve(args.lc)
        elif args.ctrate is not None:
            tstart = assign_value_if_none(args.tstart, 0)
            tstop = assign_value_if_none(args.tstop, 1025)
            t = np.arange(tstart, tstop)
            lc = Lightcurve(time=t, counts=args.ctrate + np.zeros_like(t))
        event_list.simulate_times(lc)
        nevents = len(event_list.time)
        event_list.pi = np.zeros(nevents, dtype=int)
        print('{} events generated'.format(nevents))
    else:
        event_list = None

    if args.deadtime is not None and event_list is not None:
        deadtime = args.deadtime[0]
        deadtime_sigma = None
        if len(args.deadtime) > 1:
            deadtime_sigma = args.deadtime[1]
        print(deadtime, deadtime_sigma)
        event_list, info = filter_for_deadtime(event_list, deadtime,
                                               dt_sigma=deadtime_sigma,
                                               return_all=True)
        print('{} events after filter'.format(len(event_list.time)))

        prior = np.zeros_like(event_list.time)

        prior[1:] = np.diff(event_list.time) - info.deadtime[:-1]

        additional_columns["PRIOR"] = {"data": prior, "format": "D"}
        additional_columns["KIND"] = {"data": info.is_event, "format": "L"}
        livetime = np.sum(prior)

    generate_fake_fits_observation(event_list=event_list,
                                   filename=args.outname,
                                   instr=args.instrument, mission=args.mission,
                                   tstart=args.tstart,
                                   tstop=args.tstop,
                                   mjdref=args.mjdref, livetime=livetime,
                                   additional_columns=additional_columns)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, print_function
from stingray.lightcurve import Lightcurve
import numpy as np
import os
from hendrics.read_events import treat_event_file
from hendrics.io import HEN_FILE_EXTENSION, get_file_type
from hendrics.lcurve import lcurve_from_events
import hendrics as hen
import glob


class TestLcurve():
    """Real unit tests."""
    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, 'data')
        cls.fits_fileA = os.path.join(cls.datadir, 'monol_testA.evt')
        cls.new_filename = \
            os.path.join(cls.datadir,
                         'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION)
        cls.calib_filename = \
            os.path.join(cls.datadir,
                         'monol_testA_nustar_fpma_ev_calib' +
                         HEN_FILE_EXTENSION)

    def test_treat_event_file_nustar(self):
        from astropy.io.fits import Header
        treat_event_file(self.fits_fileA)
        lcurve_from_events(self.new_filename)
        newfile = \
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_lc' + HEN_FILE_EXTENSION)
        assert os.path.exists(newfile)
        type, data = get_file_type(newfile)
        assert type == 'lc'
        assert isinstance(data, Lightcurve)
        Header.fromstring(data.header)
        assert hasattr(data, 'mjdref')
        assert data.mjdref > 0

    def test_treat_event_file_nustar_energy(self):
        command = '{0} -r {1} --nproc 2'.format(
            self.new_filename,
            os.path.join(self.datadir, 'test.rmf'))
        hen.calibrate.main(command.split())
        lcurve_from_events(self.calib_filename, e_interval=[3, 50])

        newfile = \
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_E3-50_lc' +
                         HEN_FILE_EXTENSION)
        assert os.path.exists(newfile)
        type, data = get_file_type(newfile)
        assert type == 'lc'
        assert isinstance(data, Lightcurve)
        assert hasattr(data, 'mjdref')
        assert data.mjdref > 0



from __future__ import division, print_function
import logging
import os
import copy
from .io import load_model, load_pds, save_model, save_pds, HEN_FILE_EXTENSION

import numpy as np
from stingray.modeling import fit_powerspectrum


def main_model(args=None):
    """Main function called by the `HENfspec` command line script."""
    import argparse
    description = ('Fit frequency spectra (PDS, CPDS, cospectrum) '
                   'with user-defined models')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs='+')
    parser.add_argument("-m", "--modelfile", type=str,
                        help="File containing an Astropy model with or without"
                             " constraints")
    parser.add_argument('--fitmethod', type=str, default="L-BFGS-B",
                        help='Any scipy-compatible fit method')
    parser.add_argument('--frequency-interval', type=float, nargs='+',
                        default=None,
                        help='Select frequency interval(s) to fit. Must be '
                             'an even number of frequencies in Hz, like '
                             '"--frequency-interval 0 2" or '
                             '"--frequency-interval 0 2 5 10", meaning that '
                             'the spectrum will be fitted between 0 and 2 Hz, '
                             'or using the intervals 0-2 Hz and 5-10 Hz.')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    freqs = args.frequency_interval
    if freqs is not None and len(freqs) % 2 != 0:
        raise ValueError("Invalid number of frequencies specified")

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENmodel.log', level=numeric_level,
                        filemode='w')

    model, kind, constraints = load_model(args.modelfile)
    if kind != 'Astropy':
        raise TypeError('At the moment, only Astropy models are accepted')

    for f in args.files:
        root = os.path.splitext(f)[0]
        spectrum = load_pds(f)

        if freqs is not None:
            good = np.zeros(len(spectrum.freq), dtype=bool)
            for f0, f1 in zip(freqs[::2], freqs[1::2]):
                local_good = (spectrum.freq >= f0) & (spectrum.freq < f1)
                good[local_good] = True

            spectrum_filt = copy.copy(spectrum)
            spectrum_filt.power = spectrum.power[good]
            spectrum_filt.freq = spectrum.freq[good]
            spectrum_filt.power_err = spectrum.power_err[good]

        priors = None
        max_post = False

        if constraints is not None and 'priors' in constraints:
            priors = constraints['priors']
            max_post = True

        parest, res = fit_powerspectrum(spectrum, model, model.parameters,
                                        max_post=max_post, priors=priors,
                                        fitmethod=args.fitmethod)

        save_model(res.model, root + '_bestfit.p')
        spectrum.best_fits = [res.model]
        print('Best-fit model:')
        print(res.model)
        save_pds(spectrum, root + '_fit' + HEN_FILE_EXTENSION)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to calculate colors and hardness."""

from __future__ import (absolute_import, division,
                        print_function)

from .io import HEN_FILE_EXTENSION, load_lcurve, save_lcurve
from .base import hen_root
import logging
from .lcurve import main as henlcurve
from stingray.lightcurve import Lightcurve
import numpy as np
import os


def colors():
    pass


def main(args=None):
    """Main function called by the `HENcolors` command line script."""
    import argparse
    description = \
        'Calculate color light curves'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("-e", "--energies", nargs=4, required=True,
                        type=str, default=None,
                        help="The energy boundaries in keV used to calculate "
                             "the color. E.g. -e 2 3 4 6 means that the "
                             "color will be calculated as 4.-6./2.-3. keV. "
                             "If --use-pi is specified, these are interpreted "
                             "as PI channels")
    parser.add_argument("-b", "--bintime", type=str, default='100',
                        help="Bin time; if negative, negative power of 2")
    parser.add_argument("-o", "--out", type=str,
                        default=None,
                        help='Output file')
    parser.add_argument('--use-pi', type=bool, default=False,
                        help="Use the PI channel instead of energies")
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENcolors.log', level=numeric_level,
                        filemode='w')

    option = '--e-interval'
    if args.use_pi:
        option = '--pi-interval'

    for f in files:
        henlcurve([f] + [option] + args.energies[:2] +
                  ['-b', args.bintime, '-d', '.', '-o',
                   'lc0' + HEN_FILE_EXTENSION])
        lc0 = load_lcurve('lc0' + HEN_FILE_EXTENSION)
        henlcurve([f] + [option] + args.energies[2:] +
                  ['-b', args.bintime, '-d', '.', '-o',
                   'lc1' + HEN_FILE_EXTENSION])
        lc1 = load_lcurve('lc1' + HEN_FILE_EXTENSION)

        time = lc0.time
        counts = lc1.countrate / lc0.countrate
        counts_err = np.sqrt(lc1.countrate_err ** 2 + lc0.countrate_err ** 2)
        scolor = Lightcurve(time=time, counts=counts, err=counts_err,
                            input_counts=False, err_dist='gauss',
                            gti=lc0.gti)
        del lc0
        del lc1
        os.unlink('lc0' + HEN_FILE_EXTENSION)
        os.unlink('lc1' + HEN_FILE_EXTENSION)

        if args.out is None:
            label = '_E_'
            if args.use_pi:
                label = '_PI_'
            label += '{3}-{2}_over_{1}-{0}'.format(*args.energies)
            args.out = hen_root(f) + label + HEN_FILE_EXTENSION
        scolor.e_intervals = np.asarray([float(k) for k in args.energies])
        scolor.use_pi = args.use_pi
        save_lcurve(scolor, args.out, lctype='Color')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to perform input/output operations."""
from __future__ import (absolute_import, division,
                        print_function)

import logging
import warnings
from stingray.gti import cross_gtis
from stingray.events import EventList
from stingray.lightcurve import Lightcurve
from stingray.powerspectrum import Powerspectrum, AveragedPowerspectrum
from stingray.crossspectrum import Crossspectrum, AveragedCrossspectrum
import sys
from stingray.pulse.modeling import SincSquareModel, sinc_square_model
try:
    import netCDF4 as nc
    HEN_FILE_EXTENSION = '.nc'
    HAS_NETCDF = True
except ImportError:
    msg = "Warning! NetCDF is not available. Using pickle format."
    logging.warning(msg)
    HEN_FILE_EXTENSION = '.p'
    HAS_NETCDF = False
    pass

try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

import collections
import numpy as np
import os.path
from .base import _order_list_of_arrays, _empty, is_string
from stingray.utils import assign_value_if_none
import os
import glob
import copy
from astropy.modeling.core import Model
import collections
import importlib

try:
    _ = np.complex256
    HAS_C256 = True
except:
    HAS_C256 = False

cpl128 = np.dtype([(str('real'), np.double),
                   (str('imag'), np.double)])
if HAS_C256:
    cpl256 = np.dtype([(str('real'), np.longdouble),
                       (str('imag'), np.longdouble)])


class EFPeriodogram(object):
    def __init__(self, freq=None, stat=None, kind=None, nbin=None, N=None,
                 peaks=None, peak_stat=None, best_fits=None, fdots=0):
        self.freq = freq
        self.stat = stat
        self.kind = kind
        self.nbin = nbin
        self.N = N
        self.peaks = peaks
        self.peak_stat = peak_stat
        self.best_fits = best_fits
        self.fdots = fdots


def _get_key(dict_like, key):
    try:
        return dict_like[key]
    except:
        return ""


def high_precision_keyword_read(hdr, keyword):
    """Read FITS header keywords, also if split in two.

    In the case where the keyword is split in two, like

        MJDREF = MJDREFI + MJDREFF

    in some missions, this function returns the summed value. Otherwise, the
    content of the single keyword

    Parameters
    ----------
    hdr : dict_like
        The header structure, or a dictionary
    keyword : str
        The key to read in the header

    Returns
    -------
    value : long double
        The value of the key, or None if keyword not present

    """
    try:
        value = np.longdouble(hdr[keyword])
        return value
    except:
        pass
    try:
        if len(keyword) == 8:
            keyword = keyword[:7]
        value = np.longdouble(hdr[keyword + 'I'])
        value += np.longdouble(hdr[keyword + 'F'])
        return value
    except:
        return None


def get_file_extension(fname):
    """Get the file extension."""
    return os.path.splitext(fname)[1]


def get_file_format(fname):
    """Decide the file format of the file."""
    ext = get_file_extension(fname)
    if ext == '.p':
        return 'pickle'
    elif ext == '.nc':
        return 'nc'
    elif ext in ['.evt', '.fits']:
        return 'fits'
    elif ext in ['.txt', '.qdp', '.csv']:
        return 'text'
    else:
        raise Exception("File format not recognized")


# ---- Base function to save NetCDF4 files
def save_as_netcdf(vars, varnames, formats, fname):
    """Save variables in a NetCDF4 file."""
    rootgrp = nc.Dataset(fname, 'w',
                         format='NETCDF4')

    for iv, v in enumerate(vars):
        dims = {}
        dimname = varnames[iv]+"dim"
        dimspec = (varnames[iv]+"dim", )

        if formats[iv] == 'c32':
            # Too complicated. Let's decrease precision
            warnings.warn("complex256 yet unsupported")
            formats[iv] = 'c16'

        if formats[iv] == 'c16':
            # unicode_literals breaks something, I need to specify str.
            if 'cpl128' not in rootgrp.cmptypes.keys():
                complex128_t = rootgrp.createCompoundType(cpl128, 'cpl128')
            vcomp = np.empty(v.shape, dtype=cpl128)
            vcomp['real'] = v.real.astype(np.float64)
            vcomp['imag'] = v.imag.astype(np.float64)
            v = vcomp
            formats[iv] = complex128_t

        unsized = False
        try:
            len(v)
        except TypeError:
            unsized = True

        if isinstance(v, collections.Iterable) and formats[iv] != str \
                and not unsized:
            dim = len(v)
            dims[dimname] = dim

            if isinstance(v[0], collections.Iterable):
                dim = len(v[0])
                dims[dimname + '_2'] = dim
                dimspec = (dimname, dimname + '_2')
        else:
            dims[dimname] = 1
            dim = 1

        for dimname in dims.keys():
            rootgrp.createDimension(dimname, dims[dimname])

        vnc = rootgrp.createVariable(varnames[iv], formats[iv],
                                     dimspec)
        try:
            if formats[iv] == str:
                vnc[0] = v
            else:
                vnc[:] = v
        except:
            print("Error:", varnames[iv], formats[iv], dimspec, v)
            raise
    rootgrp.close()


def read_from_netcdf(fname):
    """Read from a netCDF4 file."""
    rootgrp = nc.Dataset(fname)
    out = {}
    for k in rootgrp.variables.keys():
        dum = rootgrp.variables[k]
        values = dum.__array__()
        # Handle special case of complex
        if dum.dtype == cpl128:
            arr = np.empty(values.shape, dtype=np.complex128)
            arr.real = values[str('real')]
            arr.imag = values[str('imag')]
            values = arr

        # Handle special case of complex
        if HAS_C256 and dum.dtype == cpl256:
            arr = np.empty(values.shape, dtype=np.complex256)
            arr.real = values[str('real')]
            arr.imag = values[str('imag')]
            values = arr

        if dum.dtype == str or dum.size == 1:
            to_save = values[0]
        else:
            to_save = values
        out[k] = to_save

    rootgrp.close()

    return out


def _dum(x):
    return x


# ----- Functions to handle file types
def get_file_type(fname, raw_data=False):
    """Return the file type and its contents.

    Only works for hendrics-format pickle or netcdf files.
    """
    contents = load_data(fname)

    keys = list(contents.keys())

    ftype_raw = contents['__sr__class__type__']
    if 'Lightcurve' in ftype_raw:
        ftype = 'lc'
        fun = load_lcurve
    elif 'Color' in ftype_raw:
        ftype = 'color'
        fun = load_lcurve
    elif 'EventList' in ftype_raw:
        ftype = 'events'
        fun = load_events
    elif 'Crossspectrum' in ftype_raw:
        ftype = 'cpds'
        fun = load_pds
    elif 'Powerspectrum' in ftype_raw:
        ftype = 'pds'
        fun = load_pds
    elif 'gti' in ftype_raw:
        ftype = 'gti'
        fun = _dum
    elif 'EFPeriodogram' in ftype_raw:
        ftype = 'folding'
        fun = load_folding
    else:
        raise ValueError('File format not understood')

    if not raw_data:
        contents = fun(fname)

    return ftype, contents


# ----- functions to save and load EVENT data
def save_events(eventlist, fname):
    """Save events in a file.

    Parameters
    ----------
    eventlist: :class:`stingray.EventList` object
        Event list to be saved
    fname: str
        Name of output file
    """
    out = {'time': eventlist.time,
           'gti': eventlist.gti,
           'pi': eventlist.pi,
           'mjdref': eventlist.mjdref,
           'tstart': np.min(eventlist.gti),
           'tstop': np.max(eventlist.gti)
           }

    out['__sr__class__type__'] = str(type(eventlist))

    if hasattr(eventlist, 'instr') and eventlist.instr is not None:
        out["instr"] = eventlist.instr
    else:
        out["instr"] = 'unknown'
    if hasattr(eventlist, 'energy') and eventlist.energy is not None:
        out['energy'] = eventlist.energy
    if hasattr(eventlist, 'header') and eventlist.header is not None:
        out["header"] = eventlist.header

    if get_file_format(fname) == 'pickle':
        _save_data_pickle(out, fname)
    elif get_file_format(fname) == 'nc':
        _save_data_nc(out, fname)


def load_events(fname):
    """Load events from a file."""
    if get_file_format(fname) == 'pickle':
        out = _load_data_pickle(fname)
    elif get_file_format(fname) == 'nc':
        out = _load_data_nc(fname)

    eventlist = EventList()

    eventlist.time = out['time']
    eventlist.gti = out['gti']
    eventlist.pi = out['pi']
    eventlist.mjdref = out['mjdref']
    if 'instr' in list(out.keys()):
        eventlist.instr = out["instr"]
    if 'energy' in list(out.keys()):
        eventlist.energy = out["energy"]
    if 'header' in list(out.keys()):
        eventlist.header = out["header"]
    return eventlist


# ----- functions to save and load LCURVE data
def save_lcurve(lcurve, fname, lctype='Lightcurve'):
    """Save Light curve to file

    Parameters
    ----------
    lcurve: :class:`stingray.Lightcurve` object
        Event list to be saved
    fname: str
        Name of output file
    """
    out = {}

    out['__sr__class__type__'] = str(lctype)

    out['counts'] = lcurve.counts
    out['counts_err'] = lcurve.counts_err
    out['time'] = lcurve.time
    out['dt'] = lcurve.dt
    out['gti'] = lcurve.gti
    out['err_dist'] = lcurve.err_dist
    out['mjdref'] = lcurve.mjdref
    out['tstart'] = lcurve.tstart
    out['tseg'] = lcurve.tseg
    if hasattr(lcurve, 'header'):
        out['header'] = lcurve.header
    if hasattr(lcurve, 'expo'):
        out['expo'] = lcurve.expo
    if hasattr(lcurve, 'base'):
        out['base'] = lcurve.base
    if lctype == 'Color':
        out['e_intervals'] = lcurve.e_intervals
        out['use_pi'] = int(lcurve.use_pi)
    elif hasattr(lcurve, 'e_interval'):
        out['e_interval'] = lcurve.e_interval
        out['use_pi'] = int(lcurve.use_pi)

    if hasattr(lcurve, 'instr'):
        out["instr"] = lcurve.instr
    else:
        out["instr"] = 'unknown'

    if get_file_format(fname) == 'pickle':
        return _save_data_pickle(out, fname)
    elif get_file_format(fname) == 'nc':
        return _save_data_nc(out, fname)


def load_lcurve(fname):
    """Load light curve from a file."""
    if get_file_format(fname) == 'pickle':
        data = _load_data_pickle(fname)
    elif get_file_format(fname) == 'nc':
        data = _load_data_nc(fname)

    lcurve = Lightcurve(data['time'], data['counts'], err=data['counts_err'],
                        gti=data['gti'], err_dist=data['err_dist'],
                        mjdref=data['mjdref'])

    if 'instr' in list(data.keys()):
        lcurve.instr = data["instr"]
    if 'expo' in list(data.keys()):
        lcurve.expo = data["expo"]
    if 'e_intervals' in list(data.keys()):
        lcurve.e_intervals = data["e_intervals"]
    if 'e_interval' in list(data.keys()):
        lcurve.e_interval = data["e_interval"]
    if 'use_pi' in list(data.keys()):
        lcurve.use_pi = bool(data["use_pi"])
    if 'header' in list(data.keys()):
        lcurve.header = data["header"]
    if 'base' in list(data.keys()):
        lcurve.base = data["base"]

    return lcurve


# ---- Functions to save epoch folding results
def save_folding(efperiodogram, fname):
    """Save PDS in a file."""

    outdata = copy.copy(efperiodogram.__dict__)
    outdata['__sr__class__type__'] = 'EFPeriodogram'
    if 'best_fits' in outdata:
        model_files = []
        for i, b in enumerate(efperiodogram.best_fits):
            mfile = fname.replace(HEN_FILE_EXTENSION, '__mod{}__.p'.format(i))
            save_model(b, mfile)
            model_files.append(mfile)
        outdata.pop('best_fits')

    if get_file_format(fname) == 'pickle':
        return _save_data_pickle(outdata, fname)
    elif get_file_format(fname) == 'nc':
        return _save_data_nc(outdata, fname)


def load_folding(fname):
    """Load PDS from a file."""
    if get_file_format(fname) == 'pickle':
        data = _load_data_pickle(fname)
    elif get_file_format(fname) == 'nc':
        data = _load_data_nc(fname)

    data.pop('__sr__class__type__')

    ef = EFPeriodogram()

    for key in data.keys():
        setattr(ef, key, data[key])
    modelfiles = glob.glob(fname.replace(HEN_FILE_EXTENSION, '__mod*__.p'))
    if len(modelfiles) >= 1:
        bmodels = []
        for mfile in modelfiles:
            if os.path.exists(mfile):
                bmodels.append(load_model(mfile)[0])
        ef.best_fits = bmodels
    if len(np.asarray(ef.peaks).shape) == 0:
        ef.peaks = [ef.peaks]
    return ef


# ---- Functions to save PDSs
def save_pds(cpds, fname, save_all=True):
    """Save PDS in a file."""
    from .base import mkdir_p

    outdata = copy.copy(cpds.__dict__)
    outdata['__sr__class__type__'] = str(type(cpds))

    if not hasattr(cpds, 'instr'):
        outdata["instr"] = 'unknown'

    outdir = fname.replace(HEN_FILE_EXTENSION, "")
    if save_all:
        mkdir_p(outdir)

    if 'lc1' in outdata:
        if save_all:
            save_lcurve(cpds.lc1,
                        os.path.join(outdir,
                                     '__lc1__' + HEN_FILE_EXTENSION))
        outdata.pop('lc1')
    if 'lc2' in outdata:
        if save_all:
            save_lcurve(cpds.lc2,
                        os.path.join(outdir,
                                     '__lc2__' + HEN_FILE_EXTENSION))
        outdata.pop('lc2')
    if 'pds1' in outdata:
        if save_all:
            save_pds(cpds.pds1,
                     os.path.join(outdir,
                                  '__pds1__' + HEN_FILE_EXTENSION))
        outdata.pop('pds1')
    if 'pds2' in outdata:
        if save_all:
            save_pds(cpds.pds2,
                     os.path.join(outdir,
                                  '__pds2__' + HEN_FILE_EXTENSION))
        outdata.pop('pds2')
    if 'cs_all' in outdata:
        if save_all:
            for i, c in enumerate(cpds.cs_all):
                save_pds(
                    c,
                    os.path.join(outdir,
                                 '__cs__{}__'.format(i) + HEN_FILE_EXTENSION))
        outdata.pop('cs_all')

    if 'best_fits' in outdata and cpds.best_fits is not None:
        model_files = []
        for i, b in enumerate(cpds.best_fits):
            mfile = \
                os.path.join(outdir,
                             fname.replace(HEN_FILE_EXTENSION,
                                           '__mod{}__.p'.format(i)))
            save_model(b, mfile)
            model_files.append(mfile)
        outdata.pop('best_fits')

    if get_file_format(fname) == 'pickle':
        return _save_data_pickle(outdata, fname)
    elif get_file_format(fname) == 'nc':
        return _save_data_nc(outdata, fname)


def load_pds(fname, nosub=False):
    """Load PDS from a file."""
    if get_file_format(fname) == 'pickle':
        data = _load_data_pickle(fname)
    elif get_file_format(fname) == 'nc':
        data = _load_data_nc(fname)

    type_string = data['__sr__class__type__']
    if 'AveragedPowerspectrum' in type_string:
        cpds = AveragedPowerspectrum()
    elif 'Powerspectrum' in type_string:
        cpds = Powerspectrum()
    elif 'AveragedCrossspectrum' in type_string:
        cpds = AveragedCrossspectrum()
    elif 'Crossspectrum' in type_string:
        cpds = Crossspectrum()
    else:
        raise ValueError('Unrecognized data type in file')

    data.pop('__sr__class__type__')
    for key in data.keys():
        setattr(cpds, key, data[key])

    outdir = fname.replace(HEN_FILE_EXTENSION, "")
    modelfiles = glob.glob(
        os.path.join(outdir, fname.replace(HEN_FILE_EXTENSION, '__mod*__.p')))
    cpds.best_fits = None
    if len(modelfiles) >= 1:
        bmodels = []
        for mfile in modelfiles:
            if os.path.exists(mfile):
                bmodels.append(load_model(mfile)[0])
        cpds.best_fits = bmodels

    if nosub:
        return cpds

    lc1_name = os.path.join(outdir, '__lc1__' + HEN_FILE_EXTENSION)
    lc2_name = os.path.join(outdir, '__lc2__' + HEN_FILE_EXTENSION)
    pds1_name = os.path.join(outdir, '__pds1__' + HEN_FILE_EXTENSION)
    pds2_name = os.path.join(outdir, '__pds2__' + HEN_FILE_EXTENSION)
    cs_all_names = glob.glob(
        os.path.join(outdir, '__cs__[0-9]__' + HEN_FILE_EXTENSION))

    if os.path.exists(lc1_name):
        cpds.lc1 = load_lcurve(lc1_name)
    if os.path.exists(lc2_name):
        cpds.lc2 = load_lcurve(lc2_name)
    if os.path.exists(pds1_name):
        cpds.pds1 = load_pds(pds1_name)
    if os.path.exists(pds2_name):
        cpds.pds2 = load_pds(pds2_name)
    if len(cs_all_names) > 0:
        cs_all = []
        for c in cs_all_names:
            cs_all.append(load_pds(c))
        cpds.cs_all = cs_all

    return cpds


# ---- GENERIC function to save stuff.
def _load_data_pickle(fname, kind="data"):
    """Load generic data in pickle format."""
    logging.info('Loading %s and info from %s' % (kind, fname))
    try:
        with open(fname, 'rb') as fobj:
            result = pickle.load(fobj)
        return result
    except Exception as e:
        raise Exception("{0} failed ({1}: {2})".format('_load_data_pickle',
                                                       type(e), e))


def _save_data_pickle(struct, fname, kind="data"):
    """Save generic data in pickle format."""
    logging.info('Saving %s and info to %s' % (kind, fname))
    try:
        with open(fname, 'wb') as fobj:
            pickle.dump(struct, fobj)
    except Exception as e:
        raise Exception("{0} failed ({1}: {2})".format('_save_data_pickle',
                                                       type(e), e))
    return


def _load_data_nc(fname):
    """Load generic data in netcdf format."""
    contents = read_from_netcdf(fname)
    keys = list(contents.keys())

    keys_to_delete = []
    for k in keys:
        if k in keys_to_delete:
            continue

        if str(contents[k]) == str('__hen__None__type__'):
            contents[k] = None

        if k[-2:] in ['_I', '_L', '_F', '_k']:
            kcorr = k[:-2]

            integer_key = kcorr + '_I'
            float_key = kcorr + '_F'
            kind_key = kcorr + '_k'
            log10_key = kcorr + '_L'

            if not (integer_key in keys and float_key in keys):
                continue
            # Maintain compatibility with old-style files:
            if not (kind_key in keys and log10_key in keys):
                contents[kind_key] = "longdouble"
                contents[log10_key] = 0

            keys_to_delete.extend([integer_key, float_key])
            keys_to_delete.extend([kind_key, log10_key])

            if contents[kind_key] == 'longdouble':
                dtype = np.longdouble
            elif contents[kind_key] == 'double':
                dtype = np.double
            else:
                raise ValueError(contents[kind_key] +
                                 ": unrecognized kind string")

            log10_part = contents[log10_key]
            if isinstance(contents[integer_key], collections.Iterable):
                integer_part = np.array(contents[integer_key], dtype=dtype)
                float_part = np.array(contents[float_key], dtype=dtype)
            else:
                integer_part = dtype(contents[integer_key])
                float_part = dtype(contents[float_key])

            contents[kcorr] = (integer_part + float_part) * 10. ** log10_part

    for k in keys_to_delete:
        del contents[k]

    return contents


def _split_high_precision_number(varname, var, probesize):
    var_log10 = 0
    if probesize == 8:
        kind_str = 'double'
    if probesize == 16:
        kind_str = 'longdouble'

    if isinstance(var, collections.Iterable):
        dum = np.min(np.abs(var))
        if dum < 1 and dum > 0.:
            var_log10 = np.floor(np.log10(dum))

        var = np.asarray(var) / (10. ** var_log10)
        var_I = np.floor(var).astype(np.long)
        var_F = np.array(var - var_I, dtype=np.double)
    else:
        if np.abs(var) < 1 and np.abs(var) > 0.:
            var_log10 = np.floor(np.log10(np.abs(var)))

        var = np.asarray(var) / 10. ** var_log10
        var_I = np.long(np.floor(var))
        var_F = np.double(var - var_I)
    return var_I, var_F, var_log10, kind_str


def _save_data_nc(struct, fname, kind="data"):
    """Save generic data in netcdf format."""
    logging.info('Saving %s and info to %s' % (kind, fname))
    varnames = []
    values = []
    formats = []

    for k in struct.keys():
        var = struct[k]

        probe = var
        if isinstance(var, collections.Iterable):
            try:
                probe = var[0]
            except:
                logging.error('This failed: %s %s in file %s' %
                              (k, repr(var), fname))
                raise Exception('This failed: %s %s in file %s' %
                                (k, repr(var), fname))

        if is_string(var):
            probekind = str
            probesize = -1
        elif var is None:
            probekind = None
        else:
            probekind = np.result_type(probe).kind
            probesize = np.result_type(probe).itemsize

        if probekind == 'f' and probesize >= 8:
            # If a (long)double, split it in integer + floating part.
            # If the number is below zero, also use a logarithm of 10 before
            # that, so that we don't lose precision
            var_I, var_F, var_log10, kind_str = \
                _split_high_precision_number(k, var, probesize)
            values.extend([var_I, var_log10, var_F, kind_str])
            formats.extend(['i8', 'i8', 'f8', str])
            varnames.extend([k + '_I', k + '_L', k + '_F', k + '_k'])
        elif probekind == str:
            values.append(var)
            formats.append(probekind)
            varnames.append(k)
        elif probekind is None:
            values.append('__hen__None__type__')
            formats.append(str)
            varnames.append(k)
        else:
            values.append(var)
            formats.append(probekind + '%d' % probesize)
            varnames.append(k)

    save_as_netcdf(values, varnames, formats, fname)


def save_data(struct, fname, ftype='data'):
    """Save generic data in hendrics format."""
    if get_file_format(fname) == 'pickle':
        _save_data_pickle(struct, fname)
    elif get_file_format(fname) == 'nc':
        _save_data_nc(struct, fname)


def load_data(fname):
    """Load generic data in hendrics format."""
    if get_file_format(fname) == 'pickle':
        return _load_data_pickle(fname)
    elif get_file_format(fname) == 'nc':
        return _load_data_nc(fname)
    else:
        raise TypeError('The file type is not recognized. Did you convert the'
                        ' original files into HENDRICS format (e.g. with '
                        'HENreadevents or HENlcurve)?')


# QDP format is often used in FTOOLS
def save_as_qdp(arrays, errors=None, filename="out.qdp", mode='w'):
    """Save arrays in a QDP file.

    Saves an array of variables, and possibly their errors, to a QDP file.

    Parameters
    ----------
    arrays: [array1, array2]
        List of variables. All variables must be arrays and of the same length.
    errors: [array1, array2]
        List of errors. The order has to be the same of arrays; the value can
        be:
        - None if no error is assigned
        - an array of same length of variable for symmetric errors
        - an array of len-2 lists for non-symmetric errors (e.g.
        [[errm1, errp1], [errm2, errp2], [errm3, errp3], ...])

    Other parameters
    ----------------
    mode : str
        the file access mode, to be passed to the open() function. Can be 'w'
        or 'a'
     """
    import numpy as np
    errors = assign_value_if_none(errors, [None for i in arrays])

    data_to_write = []
    list_of_errs = []
    for ia, ar in enumerate(arrays):
        data_to_write.append(ar)
        if errors[ia] is None:
            continue
        shape = np.shape(errors[ia])
        assert shape[0] == len(ar), \
            'Errors and arrays must have same length'
        if len(shape) == 1:
            list_of_errs.append([ia, 'S'])
            data_to_write.append(errors[ia])
        elif shape[1] == 2:
            list_of_errs.append([ia, 'T'])
            mine = [k[0] for k in errors[ia]]
            maxe = [k[1] for k in errors[ia]]
            data_to_write.append(mine)
            data_to_write.append(maxe)

    print_header = True
    if os.path.exists(filename) and mode == 'a':
        print_header = False
    outfile = open(filename, mode)
    if print_header:
        for l in list_of_errs:
            i, kind = l
            print('READ %s' % kind + 'ERR %d' % (i + 1), file=outfile)

    length = len(data_to_write[0])
    for i in range(length):
        for idw, d in enumerate(data_to_write):
            print(d[i], file=outfile, end=" ")
        print("", file=outfile)

    outfile.close()


def save_as_ascii(cols, filename="out.txt", colnames=None,
                  append=False):
    """Save arrays as TXT file with respective errors."""
    import numpy as np

    logging.debug('%s %s' % (repr(cols), repr(np.shape(cols))))
    if append:
        txtfile = open(filename, "a")
    else:
        txtfile = open(filename, "w")
    shape = np.shape(cols)
    ndim = len(shape)

    if ndim == 1:
        cols = [cols]
    elif ndim > 3 or ndim == 0:
        logging.error("Only one- or two-dim arrays accepted")
        return -1
    lcol = len(cols[0])

    if colnames is not None:
        print("#", file=txtfile, end=' ')
        for i_c, c in enumerate(cols):
            print(colnames[i_c], file=txtfile, end=' ')
        print('', file=txtfile)
    for i in range(lcol):
        for c in cols:
            print(c[i], file=txtfile, end=' ')

        print('', file=txtfile)
    txtfile.close()
    return 0


def print_fits_info(fits_file, hdu=1):
    """Print general info about an observation."""
    from astropy.io import fits as pf
    from astropy.units import Unit
    from astropy.time import Time

    lchdulist = pf.open(fits_file)

    datahdu = lchdulist[hdu]
    header = datahdu.header

    info = {}
    info['N. events'] = _get_key(header, 'NAXIS2')
    info['Telescope'] = _get_key(header, 'TELESCOP')
    info['Instrument'] = _get_key(header, 'INSTRUME')
    info['OBS_ID'] = _get_key(header, 'OBS_ID')
    info['Target'] = _get_key(header, 'OBJECT')
    info['Start'] = _get_key(header, 'DATE-OBS')
    info['Stop'] = _get_key(header, 'DATE-END')

    # Give time in MJD
    mjdref = high_precision_keyword_read(header, 'MJDREF')
    tstart = high_precision_keyword_read(header, 'TSTART')
    tstop = high_precision_keyword_read(header, 'TSTOP')
    tunit = _get_key(header, 'TIMEUNIT')
    start_mjd = Time(mjdref, format='mjd') + tstart * Unit(tunit)
    stop_mjd = Time(mjdref, format='mjd') + tstop * Unit(tunit)

    print('ObsID:         {0}\n'.format(info['OBS_ID']))
    print('Date:          {0} -- {1}\n'.format(info['Start'], info['Stop']))
    print('Date (MJD):    {0} -- {1}\n'.format(start_mjd, stop_mjd))
    print('Instrument:    {0}/{1}\n'.format(info['Telescope'],
                                            info['Instrument']))
    print('Target:        {0}\n'.format(info['Target']))
    print('N. Events:     {0}\n'.format(info['N. events']))

    lchdulist.close()
    return info


def _get_gti_from_extension(lchdulist, accepted_gtistrings=['GTI']):
    hdunames = [h.name for h in lchdulist]
    gtiextn = [ix for ix, x in enumerate(hdunames)
               if x in accepted_gtistrings][0]
    gtiext = lchdulist[gtiextn]
    gtitable = gtiext.data

    colnames = [col.name for col in gtitable.columns.columns]
    # Default: NuSTAR: START, STOP. Otherwise, try RXTE: Start, Stop
    if 'START' in colnames:
        startstr, stopstr = 'START', 'STOP'
    else:
        startstr, stopstr = 'Start', 'Stop'

    gtistart = np.array(gtitable.field(startstr), dtype=np.longdouble)
    gtistop = np.array(gtitable.field(stopstr), dtype=np.longdouble)
    gti_list = np.array([[a, b]
                         for a, b in zip(gtistart,
                                         gtistop)],
                        dtype=np.longdouble)
    return gti_list


def _get_gti_from_all_extensions(lchdulist, accepted_gtistrings=['GTI'],
                                 det_numbers=None):
    if det_numbers is None:
        return _get_gti_from_extension(lchdulist, accepted_gtistrings)

    gti_lists = []
    for i in det_numbers:
        acc_gti_str = [x + '{:02d}'.format(i) for x in accepted_gtistrings]
        acc_gti_str += [x + '{:d}'.format(i) for x in accepted_gtistrings]
        gti_lists.append(_get_gti_from_extension(lchdulist, acc_gti_str))

    return cross_gtis(gti_lists)


def _get_additional_data(lctable, additional_columns):
    additional_data = {}
    if additional_columns is not None:
        for a in additional_columns:
            try:
                additional_data[a] = np.array(lctable.field(a))
            except KeyError:
                warnings.warn("Column {} not found".format(a))
                additional_data[a] = np.zeros(len(lctable))

    return additional_data


def load_gtis(fits_file, gtistring=None):
    """Load GTI from HDU EVENTS of file fits_file."""
    from astropy.io import fits as pf
    import numpy as np

    gtistring = assign_value_if_none(gtistring, 'GTI')
    logging.info("Loading GTIS from file %s" % fits_file)
    lchdulist = pf.open(fits_file, checksum=True)
    lchdulist.verify('warn')

    gtitable = lchdulist[gtistring].data
    gti_list = np.array([[a, b]
                         for a, b in zip(gtitable.field('START'),
                                         gtitable.field('STOP'))],
                        dtype=np.longdouble)
    lchdulist.close()
    return gti_list


def _get_detector_id(lctable):
    """Multi-mission detector id finder

    Examples
    --------
    >>> from astropy.io import fits
    >>> import numpy as np
    >>> a = fits.Column(name='CCDNR', array=np.array([1, 2]), format='K')
    >>> t = fits.TableHDU.from_columns([a])
    >>> det_id1 = _get_detector_id(t.data)
    >>> a = fits.Column(name='pcuid', array=np.array([1, 2]), format='K')
    >>> t = fits.TableHDU.from_columns([a])
    >>> det_id2 = _get_detector_id(t.data)
    >>> np.all(det_id1 == det_id2)
    True
    >>> a = fits.Column(name='asdfasdf', array=np.array([1, 2]), format='K')
    >>> t = fits.TableHDU.from_columns([a])
    >>> _get_detector_id(t.data) is None
    True
    """
    for column in ['CCDNR',  # XMM
                   'ccd_id',  # Chandra
                   'PCUID']:  # XTE
        for name in lctable.columns.names:
            if column.lower() == name.lower():
                return np.array(lctable.field(name), dtype=np.int)

    return None


def load_events_and_gtis(fits_file, additional_columns=None,
                         gtistring='GTI,STDGTI',
                         gti_file=None, hduname='EVENTS', column='TIME'):
    """Load event lists and GTIs from one or more files.

    Loads event list from HDU EVENTS of file fits_file, with Good Time
    intervals. Optionally, returns additional columns of data from the same
    HDU of the events.

    Parameters
    ----------
    fits_file : str
    return_limits: bool, optional
        Return the TSTART and TSTOP keyword values
    additional_columns: list of str, optional
        A list of keys corresponding to the additional columns to extract from
        the event HDU (ex.: ['PI', 'X'])

    Returns
    -------
    ev_list : array-like
    gtis: [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
    additional_data: dict
        A dictionary, where each key is the one specified in additional_colums.
        The data are an array with the values of the specified column in the
        fits file.
    t_start : float
    t_stop : float
    """
    from astropy.io import fits as pf

    gtistring = assign_value_if_none(gtistring, 'GTI,STDGTI')
    lchdulist = pf.open(fits_file)

    # Load data table
    try:
        lctable = lchdulist[hduname].data
    except:  # pragma: no cover
        logging.warning('HDU %s not found. Trying first extension' % hduname)
        lctable = lchdulist[1].data
        hduname = 1

    # Read event list
    ev_list = np.array(lctable.field(column), dtype=np.longdouble)
    detector_id = _get_detector_id(lctable)
    det_number = None if detector_id is None else list(set(detector_id))
    header = lchdulist[1].header
    # Read TIMEZERO keyword and apply it to events
    try:
        timezero = np.longdouble(header['TIMEZERO'])
    except:  # pragma: no cover
        logging.warning("No TIMEZERO in file")
        timezero = np.longdouble(0.)

    try:
        instr = header['INSTRUME']
    except:
        instr = 'unknown'

    ev_list += timezero

    # Read TSTART, TSTOP from header
    try:
        t_start = np.longdouble(header['TSTART'])
        t_stop = np.longdouble(header['TSTOP'])
    except:  # pragma: no cover
        logging.warning("Tstart and Tstop error. using defaults")
        t_start = ev_list[0]
        t_stop = ev_list[-1]

    mjdref = np.longdouble(high_precision_keyword_read(header, 'MJDREF'))

    # Read and handle GTI extension
    accepted_gtistrings = gtistring.split(',')

    if gti_file is None:
        # Select first GTI with accepted name
        try:
            gti_list = \
                _get_gti_from_all_extensions(
                    lchdulist, accepted_gtistrings=accepted_gtistrings,
                    det_numbers=det_number)
        except:  # pragma: no cover
            warnings.warn("No extensions found with a valid name. "
                          "Please check the `accepted_gtistrings` values.")
            gti_list = np.array([[t_start, t_stop]],
                                dtype=np.longdouble)
    else:
        gti_list = load_gtis(gti_file, gtistring)

    if additional_columns is None:
        additional_columns = ['PI']
    if 'PI' not in additional_columns:
        additional_columns.append('PI')

    additional_data = _get_additional_data(lctable, additional_columns)

    lchdulist.close()

    # Sort event list
    order = np.argsort(ev_list)
    ev_list = ev_list[order]

    additional_data = _order_list_of_arrays(additional_data, order)
    pi = additional_data['PI'][order]
    additional_data.pop('PI')

    returns = _empty()

    returns.ev_list = EventList(ev_list, gti=gti_list, pi=pi)

    returns.ev_list.instr = instr
    returns.ev_list.mjdref = mjdref
    returns.ev_list.header = header.tostring()
    returns.additional_data = additional_data
    returns.t_start = t_start
    returns.t_stop = t_stop
    returns.detector_id = detector_id

    return returns


def main(args=None):
    """Main function called by the `HENreadfile` command line script."""
    from astropy.time import Time
    import astropy.units as u
    import argparse

    description = \
        'Print the content of HENDRICS files'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')

    args = parser.parse_args(args)

    for fname in args.files:
        print()
        print('-' * len(fname))
        print('{0}'.format(fname))
        print('-' * len(fname))
        if fname.endswith('.fits') or fname.endswith('.evt'):
            print('This FITS file contains:', end='\n\n')
            print_fits_info(fname)
            print('-' * len(fname))
            continue
        ftype, contents = get_file_type(fname, raw_data=True)
        print('This file contains:', end='\n\n')
        if 'mjdref' in contents:
            mjdref = Time(contents['mjdref'], format='mjd')

        for k in sorted(contents.keys()):
            if k == 'tstart':
                timeval = contents[k] * u.s
                val = '{0} (MJD {1})'.format(contents[k], mjdref + timeval)
            if k == 'tseg':
                val = '{0} s'.format(contents[k])
            else:
                val = contents[k]
            if isinstance(val, collections.Iterable) and not is_string(val):
                length = len(val)
                if len(val) < 4:
                    val = repr(list(val[:4]))
                else:
                    val = repr(list(val[:4])).replace(']', '') + '...]'
                    val = '{} (len {})'.format(val, length)
            print((k + ':').ljust(15), val, end='\n\n')

        print('-' * len(fname))


def sort_files(files):
    """Sort a list of HENDRICS files, looking at `Tstart` in each."""
    allfiles = {}
    ftypes = []

    for f in files:
        logging.info('Loading file ' + f)
        ftype, contents = get_file_type(f)
        instr = contents.instr
        ftypes.append(ftype)
        if instr not in list(allfiles.keys()):
            allfiles[instr] = []
        # Add file name to the dictionary
        contents.__sort__filename__ = f
        allfiles[instr].append(contents)

    # Check if files are all of the same kind (lcs, PDSs, ...)
    ftypes = list(set(ftypes))
    assert len(ftypes) == 1, 'Files are not all of the same kind.'

    instrs = list(allfiles.keys())
    for instr in instrs:
        contents = list(allfiles[instr])
        tstarts = [np.min(c.gti) for c in contents]
        fnames = [c.__sort__filename__ for c in contents]

        fnames = [x for (y, x) in sorted(zip(tstarts, fnames))]

        # Substitute dictionaries with the sorted list of files
        allfiles[instr] = fnames

    return allfiles


def save_model(model, fname='model.p', constraints=None):
    """Save best-fit models to data.

    Parameters
    ----------
    model : func or `astropy.modeling.core.Model` object
        The model to be saved
    fname : str, default 'models.p'
        The output file name

    Other parameters
    ----------------
    constraints: dict
        Additional model constraints. Ignored for astropy models.
    """
    modeldata = {'model': model, 'constraints': None}
    if isinstance(model, (Model, SincSquareModel)):
        modeldata['kind'] = 'Astropy'
    elif callable(model):
        nargs = model.__code__.co_argcount
        nkwargs = len(model.__defaults__)
        if not nargs - nkwargs == 1:
            raise TypeError("Accepted callable models have only one "
                            "non-keyword argument")
        modeldata['kind'] = 'callable'
        modeldata['constraints'] = constraints
    else:
        raise TypeError("The model has to be an Astropy model or a callable"
                        " with only one non-keyword argument")

    with open(fname, 'wb') as fobj:
        pickle.dump(modeldata, fobj)


def load_model(modelstring):

    if not is_string(modelstring):
        raise TypeError('modelstring has to be an existing file name')
    if not os.path.exists(modelstring):
        raise FileNotFoundError('Model file not found')

    # modelstring is a pickle file
    if modelstring.endswith('.p'):
        logging.debug('Loading model from pickle file')
        with open(modelstring, 'rb') as fobj:
            modeldata = pickle.load(fobj)
        return modeldata['model'], modeldata['kind'], modeldata['constraints']
    # modelstring is a python file
    elif modelstring.endswith('.py'):
        logging.debug('Loading model from Python source')
        modulename = modelstring.replace('.py', '')
        sys.path.append(os.getcwd())
        # If a module with the same name was already imported, unload it!
        # This is because the user might be using the same file name but
        # different models inside, just like we do in test_io.py
        if modulename in sys.modules:
            del sys.modules[modulename]

        # This invalidate_caches() is called to account for the case when
        # the model file does not exist the first time we call
        # importlib.import_module(). In this case, the second time we call it,
        # even if the file exists it will not exist for importlib.
        # Unfortunately, this does not work in Python 2.
        try:
            importlib.invalidate_caches()
        except AttributeError:
            logging.warning("importlib.invalidate_caches() is not implemented "
                            "in Python 2")

        _model = importlib.import_module(modulename)
        model = _model.model
        constraints = None
        if hasattr(_model, 'constraints'):
            constraints = _model.constraints
    else:
        raise TypeError('Unknown file type')

    if isinstance(model, Model):
        return model, 'Astropy', constraints
    elif callable(model):
        nargs = model.__code__.co_argcount
        nkwargs = len(model.__defaults__)
        if not nargs - nkwargs == 1:
            raise TypeError("Accepted callable models have only one "
                            "non-keyword argument")
        return model, 'callable', constraints



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Function to sum frequency spectra."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from .io import save_pds, get_file_type
from .io import HEN_FILE_EXTENSION
from .base import _assign_value_if_none
import numpy as np
import logging
import copy


def sum_fspec(files, outname=None):
    """Take a bunch of (C)PDSs and sums them."""
    # Read first file
    ftype0, contents = get_file_type(files[0])
    pdstype = ftype0.replace('reb', '')

    freq0 = contents.freq
    pds0 = contents.power
    epds0 = contents.power_err
    nchunks0 = contents.m
    rebin0 = 1
    norm0 = contents.norm

    tot_pds = pds0 * nchunks0
    tot_epds = epds0 ** 2 * nchunks0
    tot_npds = nchunks0
    tot_contents = copy.copy(contents)
    outname = _assign_value_if_none(outname,
                                    'tot_' + ftype0 + HEN_FILE_EXTENSION)

    for f in files[1:]:
        ftype, contents = get_file_type(f)
        pdstype = ftype.replace('reb', '')

        freq = contents.freq
        pds = contents.power
        epds = contents.power_err
        nchunks = contents.m
        rebin = 1
        norm = contents.norm
        fftlen = contents.fftlen

        assert ftype == ftype0, 'Files must all be of the same kind'
        assert np.all(rebin == rebin0), \
            'Files must be rebinned in the same way'
        np.testing.assert_array_almost_equal(
            freq, freq0, decimal=int(-np.log10(1 / fftlen) + 2),
            err_msg='Frequencies must coincide')
        assert norm == norm0, 'Files must have the same normalization'

        tot_pds += pds * nchunks
        tot_epds += epds ** 2 * nchunks ** 2
        tot_npds += nchunks

    tot_contents.power = tot_pds / tot_npds
    tot_contents.power_err = np.sqrt(tot_epds) / tot_npds
    tot_contents.m = tot_npds

    logging.info('Saving %s to %s' % (pdstype, outname))
    save_pds(tot_contents, outname)

    return tot_contents


def main(args=None):
    """Main function called by the `HENsumfspec` command line script."""
    import argparse

    description = 'Sum (C)PDSs contained in different files'
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs='+')

    parser.add_argument("-o", "--outname", type=str, default=None,
                        help='Output file name for summed (C)PDS. Default:' +
                        ' tot_(c)pds' + HEN_FILE_EXTENSION)

    args = parser.parse_args(args)

    sum_fspec(args.files, args.outname)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to rebin light curves and frequency spectra."""
from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

import numpy as np
from .io import get_file_type
from .io import save_lcurve, save_pds
from .io import HEN_FILE_EXTENSION, get_file_extension
from .base import _empty, _assign_value_if_none
import logging


def rebin_file(filename, rebin):
    """Rebin the contents of a file, be it a light curve or a spectrum."""
    ftype, contents = get_file_type(filename)

    if ftype not in ['lc', 'pds', 'cpds']:
        raise ValueError('This format does not support rebin (yet):', ftype)

    if rebin == np.int(rebin):
        contents = contents.rebin(f=rebin)
    else:
        contents = contents.rebin_log(f=rebin)

    if ftype == 'lc':
        func = save_lcurve
    elif ftype in ['pds', 'cpds']:
        func = save_pds

    outfile = filename.replace(get_file_extension(filename),
                               '_rebin%g' % rebin + HEN_FILE_EXTENSION)
    logging.info('Saving %s to %s' % (ftype, outfile))
    func(contents, outfile)


def main(args=None):
    """Main function called by the `HENrebin` command line script."""
    import argparse
    description = 'Rebin light curves and frequency spectra. '
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs='+')
    parser.add_argument("-r", "--rebin", type=float, default=1,
                        help="Rebinning to apply. Only if the quantity to" +
                        " rebin is a (C)PDS, it is possible to specify a" +
                        " non-integer rebin factor, in which case it is" +
                        " interpreted as a geometrical binning factor")

    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENrebin.log', level=numeric_level,
                        filemode='w')
    rebin = args.rebin
    for f in files:
        rebin_file(f, rebin)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""First set of tests."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

import hendrics as hen
import numpy as np
import logging
import os
import unittest
import pytest

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

logging.basicConfig(filename='HEN.log', level=logging.DEBUG, filemode='w')
curdir = os.path.abspath(os.path.dirname(__file__))
datadir = os.path.join(curdir, 'data')


def _ratio(a, b):
    return np.abs(a - b) / np.abs(a + b)


class TestAll(unittest.TestCase):
    """Real unit tests."""

    def test_common_name(self):
        """Test the common_name function."""
        a = 'A_3-50_A.nc'
        b = 'B_3-50_B.nc'
        assert hen.base.common_name(a, b) == '3-50'

    def test_exposure_calculation1(self):
        """Test if the exposure calculator works correctly."""
        times = np.array([1., 2., 3.])
        events = np.array([2.])
        priors = np.array([2.])
        dt = np.array([1., 1., 1.])
        expo = hen.exposure.get_livetime_per_bin(times, events, priors, dt=dt,
                                                 gti=None)
        np.testing.assert_almost_equal(expo, np.array([1, 0.5, 0.]))

    def test_exposure_calculation2(self):
        """Test if the exposure calculator works correctly."""
        times = np.array([1., 2.])
        events = np.array([2.1])
        priors = np.array([0.3])
        dt = np.array([1., 1.])
        expo = hen.exposure.get_livetime_per_bin(times, events, priors, dt=dt,
                                                 gti=None)
        np.testing.assert_almost_equal(expo, np.array([0, 0.3]))

    def test_exposure_calculation3(self):
        """Test if the exposure calculator works correctly."""
        times = np.array([1., 2., 3.])
        events = np.array([2.1])
        priors = np.array([0.7])
        dt = np.array([1., 1., 1.])
        expo = hen.exposure.get_livetime_per_bin(times, events, priors, dt=dt,
                                                 gti=None)
        np.testing.assert_almost_equal(expo, np.array([0.1, 0.6, 0.]))

    def test_exposure_calculation4(self):
        """Test if the exposure calculator works correctly."""
        times = np.array([1., 1.5, 2., 2.5, 3.])
        events = np.array([2.6])
        priors = np.array([1.5])
        dt = np.array([0.5, 0.5, 0.5, 0.5, 0.5])
        expected_expo = np.array([0.15, 0.5, 0.5, 0.35, 0])
        expo = hen.exposure.get_livetime_per_bin(times, events, priors, dt=dt,
                                                 gti=None)
        np.testing.assert_almost_equal(expo, expected_expo)

    def test_exposure_calculation5(self):
        """Test if the exposure calculator works correctly."""
        times = np.array([1., 2., 3.])
        events = np.array([1.1, 1.2, 1.4, 1.5, 1.8, 4])
        # dead time = 0.05
        priors = np.array([0.55, 0.05, 0.15, 0.05, 0.25, 2.15])
        dt = np.array([1, 1, 1])
        expected_expo = np.array([0.8, 0.9, 1])
        expo = hen.exposure.get_livetime_per_bin(times, events, priors, dt=dt,
                                                 gti=None)
        np.testing.assert_almost_equal(expo, expected_expo)

    def test_high_precision_keyword(self):
        """Test high precision FITS keyword read."""
        from hendrics.io import high_precision_keyword_read
        hdr = {"MJDTESTI": 100, "MJDTESTF": np.longdouble(0.5),
               "CIAO": np.longdouble(0.)}
        assert \
            high_precision_keyword_read(hdr,
                                        "MJDTEST") == np.longdouble(100.5), \
            "Keyword MJDTEST read incorrectly"
        assert \
            high_precision_keyword_read(hdr, "CIAO") == np.longdouble(0.), \
            "Keyword CIAO read incorrectly"

    def test_filter_for_deadtime_nonpar(self):
        """Test dead time filter, non-paralyzable case."""
        events = np.array([1, 1.05, 1.07, 1.08, 1.1, 2, 2.2, 3, 3.1, 3.2])
        filt_events = hen.fake.filter_for_deadtime(events, 0.11)
        expected = np.array([1, 2, 2.2, 3, 3.2])
        assert np.all(filt_events == expected), \
            "Wrong: {} vs {}".format(filt_events, expected)

    def test_filter_for_deadtime_nonpar_bkg(self):
        """Test dead time filter, non-paralyzable case, with background."""
        events = np.array([1.1, 2, 2.2, 3, 3.2])
        bkg_events = np.array([1, 3.1])
        filt_events, info = \
            hen.fake.filter_for_deadtime(events, 0.11, bkg_ev_list=bkg_events,
                                         return_all=True)
        expected_ev = np.array([2, 2.2, 3, 3.2])
        expected_bk = np.array([1])
        assert np.all(filt_events == expected_ev), \
            "Wrong: {} vs {}".format(filt_events, expected_ev)
        assert np.all(info.bkg == expected_bk), \
            "Wrong: {} vs {}".format(info.bkg, expected_bk)

    def test_filter_for_deadtime_par(self):
        """Test dead time filter, paralyzable case."""
        events = np.array([1, 1.1, 2, 2.2, 3, 3.1, 3.2])
        assert np.all(hen.fake.filter_for_deadtime(
            events, 0.11, paralyzable=True) == np.array([1, 2, 2.2, 3]))

    def test_filter_for_deadtime_par_bkg(self):
        """Test dead time filter, paralyzable case, with background."""
        events = np.array([1.1, 2, 2.2, 3, 3.2])
        bkg_events = np.array([1, 3.1])
        filt_events, info = \
            hen.fake.filter_for_deadtime(events, 0.11, bkg_ev_list=bkg_events,
                                         paralyzable=True, return_all=True)
        expected_ev = np.array([2, 2.2, 3])
        expected_bk = np.array([1])
        assert np.all(filt_events == expected_ev), \
            "Wrong: {} vs {}".format(filt_events, expected_ev)
        assert np.all(info.bkg == expected_bk), \
            "Wrong: {} vs {}".format(info.bkg, expected_bk)

    def test_deadtime_mask_par(self):
        """Test dead time filter, paralyzable case, with background."""
        events = np.array([1.1, 2, 2.2, 3, 3.2])
        bkg_events = np.array([1, 3.1])
        filt_events, info = \
            hen.fake.filter_for_deadtime(events, 0.11, bkg_ev_list=bkg_events,
                                         paralyzable=True, return_all=True)

        assert np.all(filt_events == events[info.mask])

    def test_deadtime_conversion(self):
        """Test the functions for count rate conversion."""
        original_rate = np.arange(1, 1000, 10)
        deadtime = 2.5e-3
        rdet = hen.base.r_det(deadtime, original_rate)
        rin = hen.base.r_in(deadtime, rdet)
        np.testing.assert_almost_equal(rin, original_rate)


if __name__ == '__main__':
    unittest.main(verbosity=2)
